# -*- coding: utf-8 -*-

# noinspection PyUnresolvedReferences
from lxml import etree as lxml_etree
import base64
import glob
import logging
import os
import time
import persistence
from datetime import datetime, timedelta
from threading import Thread, Condition, Lock
from xml.etree import cElementTree as eTree
import iso8601
import pem
from common import FiscalParameterController
from comp_exceptions import FiscalValidation
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from dateutil import tz
from fiscalinterface import FiscalProcessor
from fiscalpersistence import FiscalDataRepository
from old_helper import convert_from_localtime_to_utc, OrderTaker, remove_xml_namespace
from msgbus import MBEasyContext
from nfcebuilder import NfeBuilder, ContextKeys
from nfcebuilder.nfceutil import NfceRequest
from pos_model import OrderParser
from pos_util import SaleLineUtil
from helper import remove_accents
from requests import ConnectionError, Timeout, Response
from typing import (
    Optional,
    Tuple,
)
from persistence import Driver as DBDriver

INF_NFE = ".//infNFe"
NFE_PROC_POS = "{0}_{1}_{2}_nfe_proc_pos{3}_{4}.xml"

loggerThread = logging.getLogger("FiscalWrapperThread")
logger = logging.getLogger("FiscalWrapper")


class NfceRequestBuilder:
    NAMESPACE_NFE = "http://www.portalfiscal.inf.br/nfe"
    NAMESPACE_SOAP = "http://www.w3.org/2003/05/soap-envelope"
    NAMESPACE_AUTORIZACAO = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeAutorizacao"
    NAMESPACE_AUTORIZACAO_4 = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeAutorizacao4"
    NAMESPACE_RET_AUTORIZACAO_4 = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeRetAutorizacao4"
    NAMESPACE_CONSULTA_SITUACAO = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeConsulta2"
    NAMESPACE_CONSULTA_SITUACAO_3 = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeConsulta3"
    NAMESPACE_CONSULTA_SITUACAO_4 = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeConsultaProtocolo4"

    def __init__(self, mbcontext, initial_order_id, crt, cnpj_contribuinte, inscr_estadual,
                 c_uf, uf, c_mun_fg, mod, serie, ambiente, end_logradouro, end_numero, end_compl, bairro, municipio,
                 cep, nome_emit, csc, cid_token, qrcode_base_url, certificate_key_path, certificate_path, order_parser,
                 fiscal_parameter_controller, sale_line_util, nfe_builder, versao_ws, synchronous_mode):
        self.mbcontext = mbcontext
        self.initial_order_id = initial_order_id
        self.crt = crt
        self.cnpj_contribuinte = cnpj_contribuinte
        self.inscr_estadual = inscr_estadual
        self.c_uf = c_uf
        self.uf = uf
        self.c_mun_fg = c_mun_fg
        self.mod = mod
        self.serie = serie
        self.ambiente = ambiente
        self.end_logradouro = end_logradouro
        self.end_numero = end_numero
        self.end_compl = end_compl
        self.bairro = bairro
        self.municipio = municipio
        self.cep = cep
        self.nome_emit = nome_emit
        self.csc = csc
        self.cid_token = cid_token
        self.qrcode_base_url = qrcode_base_url
        self.certificate_key = open(certificate_key_path, "rb").read().replace("\r\n", "\n")
        self.certificate_cert = open(certificate_path, "rb").read().replace("\r\n", "\n")
        self.order_parser = order_parser  # type: OrderParser
        self.fiscal_parameter_controller = fiscal_parameter_controller  # type: FiscalParameterController
        self.sale_line_util = sale_line_util  # type: SaleLineUtil
        self.nfe_builder = nfe_builder  # type: NfeBuilder
        self.versao_ws = versao_ws  # type: int
        self.synchronous_mode = synchronous_mode

    def build_request(self, order, contingencia, dh_contingencia, just_contingencia):
        nfe, data_emissao, serie_nota, numero_nota = self._build_nfe(order, contingencia, dh_contingencia, just_contingencia)

        envi_nfe = """<enviNFe versao="%.2f" xmlns="%s"><idLote>%s</idLote><indSinc>%s</indSinc>%s</enviNFe>""" % (
            3.1 if self.versao_ws in (1, 3) else 4,
            NfceRequestBuilder.NAMESPACE_NFE,
            int(round(time.time() * 1000)),
            "1" if self.synchronous_mode else "0",
            nfe)
        if self.versao_ws in (1, 3):
            envelopado = self.envelopa(envi_nfe, "http://www.portalfiscal.inf.br/nfe/wsdl/NfeAutorizacao", self.c_uf)
        else:
            envelopado = self.envelopa(envi_nfe, self.NAMESPACE_AUTORIZACAO_4, self.c_uf, float(self.versao_ws))
        return envelopado, data_emissao, serie_nota, numero_nota

    def build_consulta(self, recibo):
        request = "<consReciNFe versao=\"%.2f\" xmlns=\"%s\"><tpAmb>%s</tpAmb><nRec>%s</nRec></consReciNFe>" % (
            3.1 if self.versao_ws in (1, 3) else 4,
            NfceRequestBuilder.NAMESPACE_NFE,
            self.ambiente,
            recibo)
        if self.versao_ws in (1, 3):
            envelopado = self.envelopa(request, "http://www.portalfiscal.inf.br/nfe/wsdl/NfeRetAutorizacao", self.c_uf)
        else:
            envelopado = self.envelopa(request, self.NAMESPACE_RET_AUTORIZACAO_4, self.c_uf)
        return envelopado

    def envelopa_lote(self, xmls):
        envi_nfe = """<enviNFe versao="%.2f" xmlns="%s"><idLote>%s</idLote><indSinc>%s</indSinc>%s</enviNFe>""" % (
            3.1 if self.versao_ws in (1, 3) else 4,
            NfceRequestBuilder.NAMESPACE_NFE,
            int(round(time.time() * 1000)),
            "1" if self.synchronous_mode else "0",
            xmls)
        if self.versao_ws in (1, 3):
            envelopado = self.envelopa(envi_nfe, "http://www.portalfiscal.inf.br/nfe/wsdl/NfeAutorizacao", self.c_uf)
        else:
            envelopado = self.envelopa(envi_nfe, self.NAMESPACE_AUTORIZACAO_4, self.c_uf, float(self.versao_ws))
        return envelopado

    def _build_nfe(self, order_xml, contingencia, dh_contingencia, just_contingencia):
        # type: (eTree.ElementTree, bool, Optional[datetime], Optional[str]) -> (str, str, unicode, int)
        try:
            order_xml = order_xml.find("Order") if order_xml.find("Order") else order_xml
            order = self.order_parser.parse_order(order_xml)
        except Exception as ex:
            raise FiscalBuildException("Error parsing order: {}".format(ex.message))

        context = {ContextKeys.is_in_contingency: contingencia, ContextKeys.contingency_datetime: dh_contingencia, ContextKeys.contingency_reason: just_contingencia}
        xml = self.nfe_builder.build_xml(order, context)
        data_emissao, serie_nota, numero_nota = [context.get(ContextKeys.data_emissao), self.serie, context.get(ContextKeys.fiscal_number)]

        logger.info("Order {} final {}. Valor {}".format(order.order_id, context.get("nfce_key")[30:], context.get("total_prod")))

        return xml, data_emissao, serie_nota, numero_nota

    @staticmethod
    def formata_data(data):
        # type: (datetime) -> str
        local_zone = tz.tzlocal()
        data = data.replace(tzinfo=local_zone)
        data_str = data.strftime("%Y-%m-%dT%H:%M:%S%z")
        data_str = data_str[:22] + ":" + data_str[22:]
        return data_str

    @staticmethod
    def envelopa(request, namespace, c_uf, versao_ws=3.1):
        prefix = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><Envelope xmlns=\"{0:s}\"><Header><nfeCabecMsg xmlns=\"{1:s}\"><cUF>{2:s}</cUF><versaoDados>{3:.2f}" \
                 "</versaoDados></nfeCabecMsg></Header><Body><nfeDadosMsg xmlns=\"{4:s}\">".format(
            NfceRequestBuilder.NAMESPACE_SOAP, namespace, c_uf, float(versao_ws), namespace)
        suffix = "</nfeDadosMsg></Body></Envelope>"
        return prefix + request + suffix


class NfceProcessor(FiscalProcessor):

    def __init__(
        self,                       # type: NfceProcessor
        mbcontext,                  # type: MBEasyContext
        nfce_request_builder,       # type: NfceRequestBuilder
        nfce_request,               # type: NfceRequest
        nfce_autorizador,           # type: NfceAutorizador
        nfce_contingencia,          # type: NfceContingencia
        url_sefaz,                  # type: str
        fiscal_sent_dir,            # type: str
        versao_ws,                  # type: int
        synchronous_mode,           # type: bool
        nfce_situation_checker,     # type: NfceSituationChecker
        xml_enviados,               # type: str
    ):
        # type: (...) -> None
        super(NfceProcessor, self).__init__()

        self.mbcontext = mbcontext
        self.nfce_request_builder = nfce_request_builder
        self.nfce_request = nfce_request
        self.nfce_autorizador = nfce_autorizador
        self.nfce_contingencia = nfce_contingencia
        self.fiscal_sent_dir = fiscal_sent_dir
        self.url_sefaz = url_sefaz
        self.versao_ws = versao_ws
        self.synchronous_mode = synchronous_mode
        self.nfce_situation_checker = nfce_situation_checker
        self.xml_enviados = xml_enviados

    def request_fiscal(self, posid, order, tenders, paf=False, fiscal_id=None):
        if self.nfce_contingencia:
            contingencia, dh_contingencia, just_contingencia = self.nfce_contingencia.is_contingencia()
        else:
            contingencia, dh_contingencia, just_contingencia = False, None, None

        order_id = order.get("orderId")
        order_id = order_id.zfill(9)

        log_message = "Nova order para processamento - Fiscal ID: {} - Order ID: {}"
        log_message = log_message.format(str(fiscal_id), str(order_id))
        logger.info(log_message)

        logger.info("Gerando XML request para order %s" % order.get("orderId"))
        request, data_emissao, serie_nota, numero_nota,  = self.nfce_request_builder.build_request(order, contingencia, dh_contingencia, just_contingencia)  # type: str
        serie_nota = str(serie_nota).zfill(3)
        numero_nota = str(numero_nota).zfill(9)
        dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[5:7], data_emissao[8:10])

        index1 = request.index("<NFe")
        index2 = request.index("</NFe>")
        request_xml = request[index1:index2 + 6]

        index1 = request.index("Id=\"NFe")
        nfe_key = request[index1 + 4:index1 + 51]

        response_consulta = None
        dir_enviados = os.path.join(self.fiscal_sent_dir, "Enviados", dir_arquivo)
        nfe_proc = None

        if not contingencia and self.nfce_autorizador:
            request_file = None
            if not os.path.exists(dir_enviados):
                os.makedirs(dir_enviados)

            try:
                logger.info("Consultando SEFAZ para autorizar Order: %s" % order_id)
                self.nfce_autorizador.logger = logger
                request, response_consulta, tentativas = self.nfce_autorizador.autoriza_notas(request)

                logger.info("Resposta da SEFAZ recebida para Order: %s" % order_id)
                logger.info("Order {} Enviada com a chave {}".format(order_id.zfill(9), nfe_key))
            except (ConnectionError, Timeout) as e:
                logger.warning("Entrando em contingencia. Motivo: %s", e)
                justificativa = "Problemas de conexao com a SEFAZ"

                contingencia, dh_contingencia, just_contingencia = self.nfce_contingencia.entra_contingencia(justificativa)
                request, data_emissao, serie_nota, numero_nota,  = self.nfce_request_builder.build_request(order, contingencia, dh_contingencia, just_contingencia)  # type: str

                index1 = request.index("<NFe")
                index2 = request.index("</NFe>")
                request_xml = request[index1:index2 + 6]

                index1 = request.index("Id=\"NFe")
                nfe_key = request[index1 + 4:index1 + 51]

                serie_nota = str(serie_nota).zfill(3)
                numero_nota = str(numero_nota).zfill(9)
                dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[5:7], data_emissao[8:10])
                dir_enviados = os.path.join(self.fiscal_sent_dir, "Enviados", dir_arquivo)
                if not os.path.exists(dir_enviados):
                    os.makedirs(dir_enviados)

                request_file = open(os.path.join(dir_enviados, "{0}_{1}_{2}_request_pos{3}_{4}_contingencia.xml".format(serie_nota, numero_nota, order_id, str(posid).zfill(2), nfe_key)), "w+")

            except Exception as ex:
                error_response_xml = ""
                if isinstance(ex, FiscalException):
                    error_response_xml = ex.response_xml

                logger.exception("Erro tratando NFCE")

                dir_erro = os.path.join(self.fiscal_sent_dir, "Erros")
                if not os.path.exists(dir_erro):
                    os.makedirs(dir_erro)

                if "protNFe" in error_response_xml:
                    protocol_xml = self._get_protocol_nfe(response_xml=error_response_xml)
                    request_xml = self._add_protocol_to_request_xml(
                        request_xml=request_xml,
                        protocol_xml=protocol_xml,
                    )

                request_file = open(os.path.join(dir_erro, "{0}_{1}_{2}_request_pos{3}_{4}.xml".format(serie_nota, numero_nota, order_id, str(posid).zfill(2), nfe_key)), "w+")
                raise ex

            finally:
                try:
                    if request_file:
                        request_file.write(request_xml)
                        request_file.close()
                except Exception as ex:
                    logger.exception("Erro salvando arquivo request pos: %s, orderid: %s, ex: %s", str(posid), order.get("orderId"), ex)

                if response_consulta is not None:
                    if self.versao_ws == 4:
                        signature_namespace = 'http://www.w3.org/2000/09/xmldsig#'
                        xml_namespaces = lxml_etree.fromstring(response_consulta).iter().next().nsmap
                        for namespace in xml_namespaces:
                            if signature_namespace == xml_namespaces[namespace]:
                                response_consulta = response_consulta.replace(namespace + ":", "")
                                response_consulta = response_consulta.replace("<Signature>", '<Signature xmlns="{}">'.format(signature_namespace))

                    try:
                        protocol_xml = self._get_protocol_nfe(response_xml=response_consulta)
                        nfe_proc = self._add_protocol_to_request_xml(
                            request_xml=request_xml,
                            protocol_xml=protocol_xml,
                        )
                        c_stat = remove_xml_namespace(response_consulta).find(".//infProt/cStat").text
                        if c_stat == "204":
                            nfe_proc = self._get_new_request_situation(
                                order_xml=order,
                                request_str=nfe_proc,
                            )

                        with open(os.path.join(dir_enviados, NFE_PROC_POS.format(serie_nota, numero_nota, order_id, str(posid).zfill(2), nfe_key)), "w+") as nfe_proc_file:
                            nfe_proc_file.write(nfe_proc)
                    except (Exception,) as _:
                        with open(os.path.join(dir_enviados, "{0}_{1}_{2}_response_pos{3}_{4}.xml".format(serie_nota, numero_nota, order_id, str(posid).zfill(2), nfe_key)), "w+") as response_file:
                            response_file.write(response_consulta)
        else:
            request, data_emissao, serie_nota, numero_nota = self.nfce_request_builder.build_request(order, contingencia, dh_contingencia, just_contingencia)  # type: str
            serie_nota = str(serie_nota).zfill(3)
            numero_nota = str(numero_nota).zfill(9)
            dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[5:7], data_emissao[8:10])
            dir_enviados = os.path.join(self.fiscal_sent_dir, "Enviados", dir_arquivo)

            if not os.path.exists(dir_enviados):
                os.makedirs(dir_enviados)

            with open(os.path.join(dir_enviados, "{0}_{1}_{2}_request_pos{3}_{4}_contingencia.xml".format(serie_nota, numero_nota, order_id, str(int(order.get("originatorId")[3:])).zfill(2), nfe_key)), "w+") as request_file:
                request_file.write(request_xml)

        emissao_date = convert_from_localtime_to_utc(iso8601.parse_date(data_emissao))

        self.order_service.set_custom_property(
            pos_id=posid,
            key="FISCALIZATION_DATE",
            value=emissao_date.strftime("%Y-%m-%dT%H:%M:%S"),
            blk_notify=True,
        )

        return request, response_consulta, nfe_proc

    def check_nf_situation(
        self,           # type: NfceProcessor
        initial_date,   # type: str
        final_date,     # type: str
    ):
        # type: (...) -> (bool, str)

        initial_date = initial_date[:4] + '-' + initial_date[4:6] + '-' + initial_date[6:]
        final_date = final_date[:4] + '-' + final_date[4:6] + '-' + final_date[6:]

        with FiscalDataRepository(self.mbcontext) as fiscal_repository:
            logger.info("Tratando Evento Consultar Situacao XML")
            conn = None
            try:
                conn = DBDriver().open(self.mbcontext, service_name="FiscalPersistence")
                offset = 0
                while True:
                    sql = """SELECT PosId, OrderId, XMLRequest, OrderPicture
                              FROM fiscal.FiscalData
                              WHERE date(DataNota, 'unixepoch', 'localtime') >= '%s'
                              AND date(DataNota, 'unixepoch', 'localtime') <= '%s'
                              ORDER BY OrderId ASC LIMIT 500 OFFSET %s""" % (initial_date, final_date, offset)
                    cursor = conn.select(sql)
                    if cursor.rows() == 0:
                        break
                    offset += cursor.rows()
                    eTree.register_namespace('', NfceRequestBuilder.NAMESPACE_NFE)

                    self._iterate_in_orders_to_check_nfe_situation(
                        cursor=cursor,
                        fiscal_repository=fiscal_repository,
                    )

            except (Exception,) as _:
                logger.exception("Erro ao Consultar Status dos XMLs")
                return False, "Erro ao Consultar Status dos XMLs"
            finally:
                if conn:
                    conn.close()

        return True, "Status Salvos com Sucesso"

    def _iterate_in_orders_to_check_nfe_situation(
        self,               # type: NfceProcessor
        cursor,             # type: persistence.Cursor
        fiscal_repository,  # type: FiscalDataRepository
    ):
        # type: (...) -> None

        for row in cursor:
            row_entry = map(row.get_entry, ("PosId", "OrderId", "XMLRequest", "OrderPicture"))
            pos_id, order_id, request, order_picture = row_entry
            logger.info("Checando situacao da order %s" % order_id)

            request_str = base64.b64decode(request + "=" * ((4 - len(request) % 4) % 4))
            req = eTree.XML(request_str)

            need_to_change_cstat_situation = self._need_to_change_cstat_situation(
                req=req,
                order_id=order_id,
            )
            if not need_to_change_cstat_situation:
                continue

            try:
                order_xml = base64.b64decode(order_picture + "=" * ((4 - len(order_picture) % 4) % 4))
                order_xml = eTree.XML(order_xml)
                order_xml = order_xml.find("Order") if order_xml.find("Order") else order_xml

                new_req = self._get_new_request_situation(
                    order_xml=order_xml,
                    request_str=request_str,
                )

                self._update_order_situation(
                    pos_id=pos_id,
                    order_id=order_id,
                    fiscal_repository=fiscal_repository,
                    new_req=new_req,
                )
            except (Exception,) as _:
                logger.exception("Erro ao buscar protocolo da order %s" % order_id)
                continue

    @staticmethod
    def _need_to_change_cstat_situation(
        req,        # type: eTree.Element
        order_id,   # type: str
    ):
        # type: (...) -> bool

        c_stat_path = "{{{0}}}protNFe/{{{0}}}infProt/{{{0}}}cStat"
        c_stat = req.find(c_stat_path.format(NfceRequestBuilder.NAMESPACE_NFE))
        if c_stat is not None:
            c_stat = c_stat.text

        if c_stat in ("100", "150"):
            logger.info("Order %s ja estava com status 100" % order_id)
            return False
        return True

    def _update_order_situation(
        self,               # type: NfceProcessor
        pos_id,             # type: str
        order_id,           # type: str
        fiscal_repository,  # type: FiscalDataRepository
        new_req,            # type: Optional[str]
    ):
        # type: (...) -> None

        if not new_req:
            logger.exception("Erro ao gerar XML da order %s" % order_id)
            fiscal_repository.set_nfce_sent(order_id, -1)
            return

        xml_base64 = base64.b64encode(new_req)
        fiscal_repository.set_nfce_sent_with_xml_bkoffice(order_id, xml_base64, 1, 0)
        fiscal_repository.set_fiscal_xml_custom_property(
            pos_id=pos_id,
            order_id=order_id,
            xml_request=xml_base64,
            blk_notify=False,
        )
        new_req_xml = remove_xml_namespace(new_req)
        inf_nfe = new_req_xml.find(INF_NFE)
        data_emissao = inf_nfe.find(".//dhEmi").text
        serie_nota = new_req_xml.find(".//serie").text
        numero_nota = new_req_xml.find(".//nNF").text.zfill(9)
        nfe_key = inf_nfe.attrib["Id"][3:]
        dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[5:7], data_emissao[8:10])
        dir_nota = os.path.join(self.fiscal_sent_dir, self.xml_enviados, dir_arquivo)
        self._create_directory_if_not_exists(dir_nota)
        nfce_file_name = NFE_PROC_POS.format(
            serie_nota, numero_nota, order_id.zfill(9), str(pos_id).zfill(2), nfe_key
        )

        for xml_file in glob.glob(dir_nota + '/*' + order_id.zfill(9) + '_nfe_proc_pos*'):
            os.remove(xml_file)

        with open(os.path.join(dir_nota, nfce_file_name), "w+") as nfe_proc_file:
            nfe_proc_file.write(new_req)

    def _get_new_request_situation(
        self,           # type: NfceProcessor
        order_xml,      # type: eTree.Element
        request_str,    # type: str
    ):
        # type: (...) -> Optional[str]

        order_id = order_xml.get("orderId")
        logger.info("Buscando protocolo para atualizar order %s" % order_id)

        req = remove_xml_namespace(request_str)
        inf_nfe = req.find(INF_NFE)
        nfe_key = inf_nfe.attrib["Id"][3:]

        c_stat, fiscal_data_xml = self._get_c_stat_from_check_situation(
            chave_xml=nfe_key,
            order_id=order_id,
        )
        if c_stat in ("100", "150", "613"):  # 100 = ok; 613 = Chave de Acesso difere da existente em BD;
            if c_stat in ("613",):
                try:
                    logger.info("Regerando XML para order %s" % order_id)
                    request = self.nfce_request_builder.build_request(
                        order=order_xml,
                        contingencia=False,
                        dh_contingencia=None,
                        just_contingencia=None,
                    )
                    envelopado = request[0]
                    logger.info("XML Regerado. Order %s " % order_id)
                    index1 = envelopado.index("<NFe")
                    index2 = envelopado.index("</NFe>")
                    new_req = envelopado[index1:index2 + 6]

                    nfe_start_tag = request_str[:request_str.index("<NFe ")]
                    nfe_end_tag = request_str[request_str.index("</NFe>") + 6:]
                    request_str = nfe_start_tag + new_req + nfe_end_tag
                    c_stat, fiscal_data_xml = self._get_c_stat_from_check_situation(
                        chave_xml=nfe_key,
                        order_id=order_id,
                    )
                    if c_stat not in ("100",):
                        raise FiscalValidation("Campo cStat diferente de 100")
                except (Exception,) as _:
                    logger.exception("Erro ao gerar XML da order %s" % order_id)
                    return None

            inf_prot_namespace = "/{{{2}}}retConsSitNFe/{{{2}}}protNFe/{{{2}}}infProt"
            inf_prot = fiscal_data_xml.find(
                (
                    "{{{0}}}Body/{{{1}}}" + "nfeResultMsg" + inf_prot_namespace).format(
                    NfceRequestBuilder.NAMESPACE_SOAP,
                    NfceRequestBuilder.NAMESPACE_CONSULTA_SITUACAO_4,
                    NfceRequestBuilder.NAMESPACE_NFE
                )
            )
            inf_prot.attrib.pop("Id", None)

            nfe_start_tag = request_str[:request_str.index("<infProt")]
            nfe_end_tag = request_str[request_str.index("</protNFe"):]
            new_req = nfe_start_tag + eTree.tostring(inf_prot) + nfe_end_tag
        else:
            raise FiscalValidation("Status diferente de 100. %s" % c_stat)

        return new_req or request_str

    @staticmethod
    def _create_directory_if_not_exists(data_path):
        try:
            if not os.path.exists(data_path):
                os.makedirs(data_path)
        except OSError:
            logger.error("Cannot create path: {}".format(data_path))

    def _get_c_stat_from_check_situation(
        self,       # type: NfceProcessor
        chave_xml,  # type: str
        order_id,   # type: str
    ):
        # type: (...) -> Tuple[str, eTree.Element]

        self.nfce_situation_checker.logger = logger
        fiscal_data = self.nfce_situation_checker.check_situation_nfe(chave_xml)
        if not fiscal_data:
            msg = "Problemas ao tentar consultar a situação da nota na sefaz. Vamos tentar novamente depois."
            raise FiscalValidation(msg)

        fiscal_data_xml = eTree.XML(fiscal_data)
        c_stat = fiscal_data_xml.find(
            ("{{{0}}}Body/{{{1}}}" + "nfeResultMsg" + "/{{{2}}}retConsSitNFe/{{{2}}}cStat").format(
                NfceRequestBuilder.NAMESPACE_SOAP,
                NfceRequestBuilder.NAMESPACE_CONSULTA_SITUACAO_4,
                NfceRequestBuilder.NAMESPACE_NFE
            )
        )
        if c_stat is None:
            raise FiscalValidation("Campo cStat Nao Encontrado")
        else:
            c_stat = c_stat.text
        logger.info("Novo protocolo recebido para a order {}. cStat: {}".format(order_id, c_stat))
        return c_stat, fiscal_data_xml

    @staticmethod
    def _add_protocol_to_request_xml(
        request_xml,    # type: str
        protocol_xml,   # type: str
    ):
        # type: (...) -> str

        request_with_protocol = "<nfeProc xmlns=\"http://www.portalfiscal.inf.br/nfe\" versao=\"4.00\">"
        request_with_protocol += request_xml + protocol_xml + "</nfeProc>"

        return request_with_protocol

    @staticmethod
    def _get_protocol_nfe(
        response_xml,  # type: str
    ):
        # type: (...) -> str

        index = response_xml.index("<protNFe")
        index2 = response_xml.index("</protNFe>")
        return response_xml[index:index2 + 10]

    def do_validation(self, get_days_to_expiration=None):
        try:
            encoded_cert = pem.parse(self.nfce_request_builder.certificate_cert)
            cert = x509.load_pem_x509_certificate(str(encoded_cert[0]), default_backend())
            if get_days_to_expiration:
                return True, str((cert.not_valid_after - datetime.now()).days)
            elif cert.not_valid_after > datetime.now():
                return True, "OK"
            else:
                return False, "$EXPIRED_STORE_CERTIFICATE"
        except Exception as _:
            logger.exception("Falha na Leitura da Validade do Certificado NFCE")
            return False, "$CERTIFICATE_EXPIRATION_READING_FAILURE"

    def terminate(self):
        if isinstance(self.nfce_contingencia, NfceContingencia):
            self.nfce_contingencia.finaliza()


class NfceStatus(object):
    NfceAutorizada = "100"
    ServicoParalisadoTemporariamento = "108"
    ServicoParalisadoSemPrevisao = "109"
    DuplicidadeNfce = "204"
    EmissaoAtrasada = "704"
    ErroInterno = "999"


class NfceAutorizador(object):
    def __init__(self, nfce_request_builder, nfce_request, url_autorizacao, url_ret_autorizacao, versao_ws, max_tentativas_envio_lote, intervalo_retentativa_lote, send_sleep_time, synchronous_mode):
        # type: (NfceRequestBuilder, NfceRequest, str, str, int, int, int, int, bool) -> NfceAutorizador
        self.nfce_request_builder = nfce_request_builder
        self.nfce_request = nfce_request
        self.url_autorizacao = url_autorizacao
        self.url_ret_autorizacao = url_ret_autorizacao
        self.versao_ws = versao_ws
        self.max_tentativas_envio_lote = max_tentativas_envio_lote
        self.intervalo_retentativa_lote = intervalo_retentativa_lote
        self.send_sleep_time = send_sleep_time
        self.synchronous_mode = synchronous_mode
        self.logger = logger

    def autoriza_notas(self, request, max_tentativas=None, envio_em_lote=False):
        # type: (str, int, bool) -> (str, str, int)

        if max_tentativas is None:
            max_tentativas = self.max_tentativas_envio_lote

        self.logger.info("Enviando lote para SEFAZ")
        if self.versao_ws not in (1, 3, 4):
            raise AttributeError("Versão WS Inválida. Parâmetros válidos: 1, 3 e 4")

        if self.synchronous_mode:
            self.logger.info("Preparando para envio em modo síncrono...")

        status_code, response = self._envia_lote_com_retentativa(request)

        if status_code == LoteStatus.RecebidoComSucesso or self.synchronous_mode:
            recibo = None
            if not self.synchronous_mode:
                self.logger.info("Lote recebido com sucesso, tentando obter o recibo...")
                recibo = self._busca_recibo(response)
                self.logger.info("Recibo obtido com sucesso")

            response_nfce, tentativas = self._busca_resposta_do_lote(recibo, max_tentativas, envio_em_lote, response)
            return request, response_nfce, tentativas
        else:
            if status_code in (LoteStatus.ServicoParalisadoTemporariamente,
                               LoteStatus.ServicoParalisadoSemPrevisao,
                               LoteStatus.ErroInterno,
                               LoteStatus.ConsumoIndevido):
                error_msg = "SEFAZ Indisponivel - Status: {}".format(status_code)
                self.logger.error(error_msg)
                raise ConnectionError(error_msg)

            error_msg = "Erro ao enviar lote de NFCe. Codigo: {}".format(status_code)
            self.logger.error(error_msg)
            raise LoteException(error_msg, status_code)

    def _envia_lote_com_retentativa(self, request):
        # type: (str) -> (str, Response)
        tentativa_atual = 0
        response = None
        while tentativa_atual < self.max_tentativas_envio_lote:
            # noinspection PyBroadException
            try:
                if self.versao_ws in (1, 3):
                    soap_act = None
                else:
                    soap_act = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeAutorizacao4/nfeAutorizacaoLote"
                response = self.nfce_request.envia_nfce(request, self.url_autorizacao, soap_act)

                if response.status_code == 200:
                    self.logger.info("Resposta do lote recebida")
                    codigo_retorno = remove_xml_namespace(response.content).find(".//retEnviNFe/cStat").text

                    return codigo_retorno, response
                else:
                    self.logger.info("Sefaz retornou status diferente de 200: {0} e content: {1}".format(response.status_code, response.content))

            except Exception as _:
                if response is not None:
                    self.logger.exception("Erro obtendo codigo retorno: StatusCode: {0}, Body: {1}".format(response.status_code, response.content))
                else:
                    self.logger.exception("Erro obtendo codigo retorno")

            tentativa_atual += 1
        else:
            self.logger.info("Falha ao enviar lote. Excedido numero de tentativas")
            return "999", None

    def _busca_recibo(self, response):
        # type: (Response) -> str

        # Lote foi recebido para processamento com sucesso, vamor pegar o recibo e inicar a consulta
        try:
            xml = remove_xml_namespace(response.content)
            return xml.find(".//infRec/nRec").text
        except Exception:
            error_msg = "Erro ao obter o recibo da NFCe"
            self.logger.exception(error_msg)
            raise FiscalException(error_msg)

    def _busca_resposta_do_lote(self, recibo, max_tentativas, envio_de_lote, resposta_lote):
        # type: (str, int, bool, Response) -> (str, int)

        processado = False
        tentativas = 0

        while not processado and (max_tentativas == -1 or tentativas < max_tentativas):
            try:
                if self.synchronous_mode:
                    response_consulta = resposta_lote
                else:
                    request_consulta = self.nfce_request_builder.build_consulta(recibo)

                    self.logger.info("Consultando status do lote na SEFAZ")
                    if self.versao_ws in (1, 3):
                        soap_act = None
                    else:
                        soap_act = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeRetAutorizacao4/nfeAutorizacaoLote"

                    time.sleep(1)  # Tempo aguardando status ser processado antes de tentar obter seu status

                    response_consulta = self.nfce_request.envia_nfce(request_consulta, self.url_ret_autorizacao, soap_act)

                if response_consulta and response_consulta.status_code == 200:
                    nfce_response_xml = eTree.XML(response_consulta.content)
                    status_consulta = self._busca_status_consulta(nfce_response_xml)
                else:
                    if not response_consulta:
                        self.logger.error("Sem resposta da sefaz")
                    else:
                        self.logger.info("Sefaz retornou status diferente de 200: {0} e content: {1}".format(response_consulta.status_code, response_consulta.content))
                    time.sleep(self.intervalo_retentativa_lote)
                    continue
            except Timeout:
                self.logger.exception("Exceção processando o lote, retentando")
                time.sleep(self.send_sleep_time)
                continue
            except Exception as _:
                self.logger.exception("Exceção processando o lote, retentando")
                time.sleep(self.intervalo_retentativa_lote)
                continue
            finally:
                tentativas += 1

            if status_consulta == LoteConsultaStatus.LoteEmProcessamento:
                self.logger.info("Lote ainda não processado. Aguardando processamento... Tentativa: %d - Status: %s" % (tentativas, status_consulta))
                time.sleep(self.intervalo_retentativa_lote)
                continue

            if status_consulta == LoteConsultaStatus.ConsumoIndevido:
                self.logger.error("Falha ao obter status da NFCe. SEFAZ Error 656 - Consumo Indevido")
                raise ConnectionError("Falha ao obter status da NFCe. SEFAZ Error 656 - Consumo Indevido")

            if status_consulta == LoteConsultaStatus.LoteProcessado:
                self.logger.info("Lote ja processado, verificando status do processamento da NFCE")
                if envio_de_lote:
                    return response_consulta.content, tentativas

                self._verifica_status_processamento_nfce(nfce_response_xml)
                return eTree.tostring(nfce_response_xml), tentativas

            if status_consulta == LoteConsultaStatus.ErroInterno:
                msg = "Erro interno não identificado ao tentar obter o status do lote na sefaz. 999 - Erro Interno"
                self.logger.error(msg)
                raise ConnectionError(msg)

            if status_consulta == LoteConsultaStatus.ServicoParalisadoTemporariamente:
                msg = "SEFAZ temporariamente indisponível. Erro 108 - Serviço Paralisado Momentaneamente"
                self.logger.error(msg)
                raise ConnectionError(msg)

            else:
                self.logger.error("Erro ao consultar status do lote NFCe. Codigo: %s", str(status_consulta))
                raise LoteException("Erro ao consultar status do lote NFCe. Codigo: %s" % str(status_consulta), status_consulta)

        self.logger.error("Falha ao obter status da NFCe. Excedido Numero de Tentativas")
        raise ConnectionError("Falha ao obter status da NFCe. Excedido Numero de Tentativas")

    def _busca_status_consulta(self, response_consulta_xml):
        # type: (eTree) -> str
        try:
            xml = remove_xml_namespace(eTree.tostring(response_consulta_xml))
            if self.synchronous_mode:
                c_stat = xml.find(".//retEnviNFe/cStat").text
            else:
                c_stat = xml.find(".//retConsReciNFe/cStat").text

            logger.info("Resposta da consulta de lote obtida")
            return c_stat

        except Exception as ex:
            logger.exception("Erro obtendo status consulta: {0}".format(eTree.tostring(response_consulta_xml)))
            raise ex

    @staticmethod
    def _verifica_status_processamento_nfce(response_consulta_xml):
        # type: (eTree) -> None
        response_consulta = eTree.tostring(response_consulta_xml)
        try:
            xml = remove_xml_namespace(response_consulta)

            status_nfce = xml.find(".//infProt/cStat").text
            x_motivo = xml.find(".//infProt/xMotivo").text if xml.find(".//infProt/xMotivo") is not None else None
        except Exception as ex:
            logger.exception("Erro obtendo status protocolo. ResponseXML: {0}".format(response_consulta))
            raise ex

        if x_motivo is None:
            x_motivo = ""
            logger.error("Nao foi possivel encontrar xMotivo: {}".format(response_consulta))

        if status_nfce == NfceStatus.NfceAutorizada:
            logger.info("Protocolo processado com sucesso, finalizado")
            return

        contingency_reasons = (
            NfceStatus.ServicoParalisadoTemporariamento,
            NfceStatus.ServicoParalisadoSemPrevisao,
            NfceStatus.ErroInterno,
            NfceStatus.EmissaoAtrasada
        )
        if status_nfce in contingency_reasons:
            logger.error("SEFAZ Indisponivel - Entrando em Contingencia - Status: %s ", status_nfce)
            raise ConnectionError("SEFAZ Indisponivel - Entrando em Contingencia - Status: %s" % status_nfce)
        elif status_nfce == NfceStatus.DuplicidadeNfce:
            logger.info("204 retornado mas mesma nota presente. Vamos aceitar")
        else:
            x_motivo = remove_accents(x_motivo)
            error_message = "NFCe Nao Autorizada. Status: {}. Motivo: {}".format(status_nfce, x_motivo)
            logger.error(error_message)
            raise FiscalException(
                message=error_message,
                motivo=x_motivo,
                response_xml=response_consulta,
            )

    @staticmethod
    def _verifica_mesma_nota(response_xml):
        # type: (eTree) -> bool
        try:
            xml = remove_xml_namespace(eTree.tostring(response_xml))
            chave_atual = xml.find(".//infProt/chNFe").text
            x_motivo = xml.find(".//infProt/xMotivo").text

            index = x_motivo.find("[chNFe: ")
            chave_original = x_motivo[index + 8:index + 8 + 44]

            return chave_atual == chave_original

        except Exception as _:
            logger.exception("Erro verificando mesma nota. Xml: " + eTree.tostring(response_xml))
            return False


class NfceConnectivityTester(object):
    def __init__(self, nfce_request, url_status_servico, c_uf, versao_ws):
        self.nfce_request = nfce_request
        self.url_status_servico = url_status_servico
        self.c_uf = c_uf
        self.versao_ws = versao_ws
        self.logger = logger
        self.last_test = None
        self.last_connection_status = False

    def test_connectivity(self):
        if self.last_test and self.last_test > datetime.now() - timedelta(minutes=3):
            return self.last_connection_status

        envi_nfe = """<enviNFe versao="%.2f" xmlns="%s"></enviNFe>""" % (3.1 if self.versao_ws in (1, 3) else 4, NfceRequestBuilder.NAMESPACE_NFE)
        namespace = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeStatusServico2" if self.versao_ws in (1, 3) else "http://www.portalfiscal.inf.br/nfe/wsdl/NFeStatusServico4"
        envelopado = NfceRequestBuilder.envelopa(envi_nfe, namespace, self.c_uf, 3.1 if self.versao_ws in (1, 3) else 4)
        soap_action = None if self.versao_ws in (1, 3) else "http://www.portalfiscal.inf.br/nfe/wsdl/NFeStatusServico4/nfeStatusServicoNF"
        has_connection = False
        try:
            resp = self.nfce_request.envia_nfce(envelopado, self.url_status_servico, soap_action)
            has_connection = resp.status_code == 200
            if has_connection:
                self.logger.info("Sucesso no teste de conexão com a sefaz")
            return has_connection
        except Exception as _:
            self.logger.exception("Falha no teste de conexão com a sefaz")
            return has_connection
        finally:
            self.last_connection_status = has_connection
            self.last_test = datetime.now()


class NfceSituationChecker(object):
    def __init__(self, nfce_request, url_consultar_situacao, ambiente, c_uf, versao_ws):
        self.nfce_request = nfce_request
        self.url_consultar_situacao = url_consultar_situacao
        self.ambiente = ambiente
        self.c_uf = c_uf
        self.versao_ws = versao_ws
        self.logger = logger

    def check_situation_nfe(self, nfe, timeout=None):
        cons_nfe =  """<consSitNFe xmlns="http://www.portalfiscal.inf.br/nfe" versao="%.2f"><tpAmb>%s</tpAmb><xServ>CONSULTAR</xServ><chNFe>%s</chNFe></consSitNFe>""" % (3.1 if self.versao_ws in (1, 3) else 4, self.ambiente, nfe)
        namespace = NfceRequestBuilder.NAMESPACE_CONSULTA_SITUACAO if self.versao_ws in (1, 3) else NfceRequestBuilder.NAMESPACE_CONSULTA_SITUACAO_4
        envelopado = NfceRequestBuilder.envelopa(cons_nfe, namespace, self.c_uf)
        soap_action = NfceRequestBuilder.NAMESPACE_CONSULTA_SITUACAO if self.versao_ws in (1, 3) else "http://www.portalfiscal.inf.br/nfe/wsdl/NFeConsultaProtocolo4/nfeConsultaNF"

        try:
            resp = self.nfce_request.envia_nfce(envelopado, self.url_consultar_situacao, soap_action, timeout)
            if resp.status_code == 200:
                return resp.text
            else:
                msg = "Erro ao consultar situação NFe - Code: {0} / Content: {1}".format(resp.status_code, resp.content)
                self.logger.error(msg)
        except Exception as _:
            msg = "Erro ao consultar situacao NFe"
            self.logger.exception(msg)


class NfceContingencia:
    def __init__(
        self,
        mbcontext,                          # type: MBEasyContext
        nfce_request_builder,               # type: NfceRequestBuilder
        nfce_autorizador,                   # type: NfceAutorizador
        versao_ws,                          # type: int
        nfce_situation_checker,             # type: NfceSituationChecker
        fiscal_sent_dir,                    # type: unicode
        max_tantavivas_envio_contingencia,  # type: int
        synchronous_mode,                   # type: bool
        nfce_connectivity_tester,           # type: NfceConnectivityTester
        database_dir,                       # type: str
        fiscal_schema_sql_path,             # type: str
    ):

        # type: (...) -> NfceContingencia

        self.mbcontext = mbcontext
        self.nfce_request_builder = nfce_request_builder
        self.nfce_aurotizador = nfce_autorizador
        self.versao_ws = versao_ws
        self.thread_condition = Condition()
        self.contingencia_lock = Lock()
        self.contingencia = False
        self.dh_contingencia = None
        self.just_contingencia = None
        self.running = True
        self.max_tentativas = 1
        self.contingencia_thread = None
        self.nfce_situation_checker = nfce_situation_checker
        self.fiscal_sent_dir = fiscal_sent_dir
        self.force_contingency = False
        self.process_now = False
        self.max_tantavivas_envio_contingencia = max_tantavivas_envio_contingencia
        self.synchronous_mode = synchronous_mode
        self.initial_batch_size = 50 if not self.synchronous_mode else 1
        self.batch_size = self.initial_batch_size
        self.wait_for_start = False
        self.nfce_connectivity_tester = nfce_connectivity_tester
        self.database_dir = database_dir
        self.fiscal_schema_sql_path = fiscal_schema_sql_path

    def start_contingencia(self):
        with FiscalDataRepository(self.mbcontext) as fiscalrepository:
            with self.contingencia_lock:
                if fiscalrepository.get_count_nfce_orders_to_send() == 0 and self.force_contingency is False:
                    self.contingencia = False
                else:
                    self.contingencia = True
                    self.dh_contingencia = datetime.now()
                    self.just_contingencia = "Notas anteriores em contingencia"

        if self.versao_ws not in (1, 3, 4):
            raise AttributeError("Versão WS Inválida. Parâmetros válidos: 1, 3 e 4")

        self.contingencia_thread = Thread(target=self._processa_contingencia)
        self.contingencia_thread.daemon = True
        self.contingencia_thread.start()

    def is_contingencia(self):
        with self.contingencia_lock:
            return self.contingencia, self.dh_contingencia, self.just_contingencia

    def entra_contingencia(self, justificativa):
        with self.contingencia_lock:
            self.contingencia = True
            self.dh_contingencia = datetime.now()
            self.just_contingencia = justificativa
            with self.thread_condition:
                loggerThread.info("Entrando em contingência...")
                self.wait_for_start = True
                self.thread_condition.notifyAll()

            return self.contingencia, self.dh_contingencia, self.just_contingencia

    def finaliza(self):
        if self.running:
            self.thread_condition.acquire()
            self.running = False
            self.thread_condition.notifyAll()
            self.thread_condition.release()

    def _processa_contingencia(self):
        count = 0
        dir_enviados = os.path.join(self.fiscal_sent_dir, "Enviados")

        while self.running:
            try:
                if self.force_contingency:
                    loggerThread.info("Contingência forçada está ativa. Não tentaremos envio a SEFAZ no momento")
                    return

                self.nfce_connectivity_tester.logger = loggerThread
                connection_working = self.nfce_connectivity_tester.test_connectivity()
                if not connection_working:
                    self._wait(300)
                    continue

                if self.process_now is False and self.contingencia is False and count < 60:
                    self.thread_condition.acquire()
                    self.thread_condition.wait(5)
                    self.thread_condition.release()

                    count += 1
                    continue

                # No release da thread, vamos aguardar para nao tratar uma order que ainda nao esta preparada
                if self.wait_for_start:
                    time.sleep(5)

                self.wait_for_start = False

                count = 0
                with FiscalDataRepository(
                    mbcontext=self.mbcontext,
                    database_dir=self.database_dir,
                    schema_sql_path=self.fiscal_schema_sql_path,
                ) as fiscalrepository:
                    orders_to_send = fiscalrepository.get_nfce_orders_to_send(self.batch_size)
                    qty_orders_selected = len(orders_to_send)

                    if self.synchronous_mode:
                        loggerThread.info("Modo síncrono ativado. Vamos processar pedidos individualmente")
                    else:
                        loggerThread.info("Encontradas {} orders para serem enviadas. Tamanho do lote: {}".format(qty_orders_selected, self.batch_size))

                    if orders_to_send:
                        all_nfces = ""

                        for order_to_send in orders_to_send[:]:
                            try:
                                xml_order_pict = eTree.fromstring(base64.b64decode(order_to_send.order_picture))
                                xml_order_pict = xml_order_pict.find(".//Order") if xml_order_pict.find(".//Order") else xml_order_pict
                                order_state = xml_order_pict.attrib["state"]
                                if order_state != "PAID":
                                    msg = "Order {} nao contém status PAID. Status da order: {}".format(order_to_send.order_id, order_state)
                                    loggerThread.error(msg)
                                    raise Exception(msg)
                            except TypeError as _:
                                loggerThread.exception("Erro tratando OrderPicture para order: {}. Vamos tentar regera-lo".format(order_to_send.order_id))

                                try:
                                    time.sleep(2)  # Vamos aguardar a order ser atualizada no banco
                                    order_pict = OrderTaker().get_order_picture(order_to_send.pos_id, order_to_send.order_id)
                                    fiscalrepository.set_order_picture(order_to_send.order_id, base64.b64encode(order_pict))
                                    loggerThread.info("Novo OrderPicture foi gerado com sucesso para order: {}".format(order_to_send.order_id))
                                    order_state = eTree.fromstring(order_pict).attrib["state"]

                                    if order_state != "PAID":
                                        loggerThread.error("Order {} nao contém status PAID. Status da order: {}. Marcando como -999".format(order_to_send.order_id, order_state))
                                        orders_to_send.remove(order_to_send)
                                        fiscalrepository.set_nfce_sent(order_to_send.order_id, -999)
                                        continue

                                except Exception as _:
                                    loggerThread.exception("Erro tentando obter novo OrderPicture! Marcando order {} como invalida.".format(order_to_send.order_id))
                                    fiscalrepository.set_nfce_sent(order_to_send.order_id, -1)
                                    orders_to_send.remove(order_to_send)
                                    continue

                            except Exception as _:
                                loggerThread.info("Order antiga e/ou sem referencia: {}".format(_.message))
                                fiscalrepository.set_nfce_sent(order_to_send.order_id, -1)
                                orders_to_send.remove(order_to_send)
                                continue

                            xml_base64 = order_to_send.xml
                            xml_string = base64.b64decode(xml_base64)
                            all_nfces += xml_string
                            order_to_send.xml = xml_string

                        if not orders_to_send or not all_nfces:
                            loggerThread.info("Nenhuma nota a ser processada. Vamos tentar obter novas notas.")
                            continue
                        else:
                            loggerThread.info("Quantidade de notas a serem processadas: {}".format(len(orders_to_send)))
                            orders_to_send = sorted(orders_to_send, key=lambda x: int(x.order_id))

                        envelopado = self.nfce_request_builder.envelopa_lote(all_nfces)
                        try:
                            self.nfce_aurotizador.logger = loggerThread
                            request, response, tentativas = self.nfce_aurotizador.autoriza_notas(envelopado, self.max_tantavivas_envio_contingencia, True)
                        except LoteException:
                            if not self.synchronous_mode and self.batch_size == self.initial_batch_size:
                                # Se estamos processando um lote inteiro e tivemos algum problema, vamos processar nota a nota
                                self.batch_size = 1
                                loggerThread.info("Problemas com o lote - Enviando XMLs individualmente")
                            else:
                                # Estamos processando uma unica nota e tivemos problemas, marcamos ela como problema e pegamos a próxima
                                fiscalrepository.set_nfce_sent(orders_to_send[0].order_id, -1)
                                loggerThread.info("Marcando XML da order {} como invalido".format(orders_to_send[0].order_id))
                                # Vamos esperar alguns segundos para enviar a proxima nota afim de nao receber 656 da SEFAZ
                                time.sleep(10)

                            # De qualquer maneira, vamos selecionar novas notas para enviar
                            continue

                        except ConnectionError:
                            loggerThread.info("Problemas de Conexao - Aguardando 5 minutos para tentar novamente")
                            self._wait(300)
                            continue

                        protocolos_notas = []
                        try:
                            response_xml = remove_xml_namespace(response)
                            loggerThread.info("Notas processadas com sucesso. Verificando protocolos")
                            protocolos_notas = response_xml.findall(".//infProt")
                        except Exception as _:
                            loggerThread.exception("Erro obtendo protocolo - Response: {}".format(response))

                        loggerThread.info("Quantidade de protocolos a serem processados: {}".format(len(protocolos_notas)))
                        for protocolo in protocolos_notas:
                            try:
                                chave_protocolo = protocolo.find(".//chNFe").text

                                for order in orders_to_send:
                                    order_id = order.order_id.zfill(9)
                                    loggerThread.info("Order a ser processada: {}".format(order_id))
                                    order_xml = remove_xml_namespace(order.xml)
                                    nfe_key = order_xml.find(INF_NFE).attrib['Id']
                                    chave_xml = nfe_key[3:]

                                    if chave_xml == chave_protocolo:
                                        data_emissao = order_xml.find(".//ide/dhEmi").text
                                        serie_nota = order_xml.find(".//ide/serie").text
                                        serie_nota = serie_nota.zfill(3)
                                        numero_nota = order_xml.find(".//ide/nNF").text
                                        numero_nota = numero_nota.zfill(9)
                                        dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[5:7], data_emissao[8:10])
                                        dir_nota = os.path.join(dir_enviados, dir_arquivo)

                                        if not os.path.exists(dir_nota):
                                            os.makedirs(dir_nota)

                                        stat = protocolo.find(".//cStat").text
                                        loggerThread.info("Protocolo encontrado para nota %s; Order: %s; Status: %s", chave_protocolo, order_id, stat)

                                        if stat in ("100", "204", "704", "539", "150", "635", "217", "558", "703"):  # 100 = ok; 204, 539 = duplicidade (ja estava ok); 150 = ok, atrasada;
                                            eTree.register_namespace('', NfceRequestBuilder.NAMESPACE_NFE)

                                            if stat in ("539", "217"):
                                                try:
                                                    loggerThread.info("Preparando pedido [{}] para ser enviada novamente na próxima tentantiva".format(order_id))
                                                    order_picture = eTree.XML(base64.b64decode(order.order_picture + "=" * ((4 - len(order.order_picture) % 4) % 4)))
                                                    envelopado, data_emissao, serie_nota, numero_nota = self.nfce_request_builder.build_request(order_picture, False, None, None)
                                                    index1 = envelopado.index("<NFe")
                                                    index2 = envelopado.index("</NFe>")
                                                    new_req = envelopado[index1:index2 + 6]
                                                    xml_base64 = base64.b64encode(new_req)
                                                    order.xml = new_req
                                                    loggerThread.info("Marcando pedido [{}] como pendente de envio".format(order_id))
                                                    fiscalrepository.set_nfce_sent_with_xml(order_id, xml_base64, 0)

                                                except FiscalBuildException as _:
                                                    loggerThread.info("Erro montando XML order: [{}] - Buscando Order Picture".format(order.order_id))
                                                    order_picture = OrderTaker().get_order_picture(order.pos_id, order.order_id)
                                                    fiscalrepository.set_order_picture(order.order_id, base64.b64encode(order_picture))

                                                except FiscalException as _:
                                                    loggerThread.info("Marcando XML nota [{}] como invalido".format(order.order_id))
                                                    fiscalrepository.set_nfce_sent(order.order_id, -1)

                                                except Exception as _:
                                                    loggerThread.exception("Erro ao gerar XML da order [{}]".format(order_id))
                                                    fiscalrepository.set_nfce_sent(order.order_id, -1)

                                                break

                                            if stat == "635":
                                                loggerThread.info("Recebemos 635, vamos esgotar as tentativas e aguardar para tentar novamente.")
                                                tentativas = self.max_tentativas + 1
                                                break

                                            if stat in ("204", "704", "558", "703"):
                                                nfe_key = order_xml.find(INF_NFE).attrib['Id']
                                                chave_xml = nfe_key[3:]

                                                # Vamos esperar alguns segundos para checar a situacao da nota afim de nao receber 656 da SEFAZ
                                                self.nfce_situation_checker.logger = loggerThread
                                                for _ in range(3):
                                                    time.sleep(5)
                                                    fiscal_data = self.nfce_situation_checker.check_situation_nfe(chave_xml, timeout=5)
                                                    if fiscal_data:
                                                        break
                                                else:
                                                    loggerThread.error("Problemas ao tentar consultar a situação da nota na sefaz. Vamos tentar novamente depois.")
                                                    break

                                                fiscal_data_xml = remove_xml_namespace(fiscal_data.encode("utf8"))
                                                c_stat = fiscal_data_xml.find(".//retConsSitNFe/cStat")

                                                if c_stat is None:
                                                    loggerThread.error("Campo cStat nao encontrado ou vazio")
                                                    break

                                                c_stat = c_stat.text

                                                loggerThread.info("Protocolo recebido. OrderId: {}; cStat: {}".format(order_id, c_stat))

                                                if c_stat in ("100", "150"):
                                                    inf_prot = fiscal_data_xml.find(".//protNFe/infProt")
                                                    inf_prot.attrib.pop("Id", None)
                                                    protocolo = inf_prot
                                                else:
                                                    fiscalrepository.set_nfce_sent(order.order_id, -1)
                                                    break

                                            nfce_wt_resp = "<nfeProc versao=\"{0:.2f}\" xmlns=\"{1}\">{2}<protNFe versao=\"{0:.2f}\">{3}</protNFe></nfeProc>"\
                                                .format(3.1 if self.versao_ws in (1, 3) else 4, NfceRequestBuilder.NAMESPACE_NFE, order.xml, eTree.tostring(protocolo))

                                            nfce_file_name = NFE_PROC_POS.format(serie_nota, numero_nota, order_id, str(order.pos_id).zfill(2), nfe_key)

                                            for xml_file in glob.glob(dir_nota + '/*_' + numero_nota + "_" + order_id.zfill(9) + '_request_pos*'):
                                                os.remove(xml_file)

                                            with open(os.path.join(dir_nota, nfce_file_name), "w+") as nfe_proc_file:
                                                nfe_proc_file.write(nfce_wt_resp)

                                            xml = nfce_wt_resp
                                            picture = self._update_order_properties(fiscalrepository, xml, order)

                                            xml_response = base64.b64encode(eTree.tostring(response_xml))
                                            fiscalrepository.set_xml_response(
                                                order_id=order.order_id,
                                                xml_base64=xml_response,
                                            )

                                            fiscalrepository.notify_pump(order_picture=picture, xml_format='Contingency')
                                        elif stat not in ("108", "109", "635", "999"):
                                            loggerThread.error("Nota com status desconhecido, vamos invalida-la. Status: {}".format(stat))
                                            fiscalrepository.set_nfce_sent(order.order_id, -1)

                                        orders_to_send.remove(order)
                                        break
                            except Exception as _:
                                loggerThread.exception("Erro tratando protocolo: {}".format(eTree.tostring(protocolo)))
                                continue

                        if tentativas > self.max_tentativas:
                            loggerThread.info("SEFAZ lenta. Não vamos sair da contingência. Aguardando 5 minutos para tentar novamente")
                            self._wait(300)
                            continue

                    # Todas as orders foram enviadas. Vamos verificar se ainda estamos em contingencia
                    if qty_orders_selected < self.batch_size and self.force_contingency is False:
                        # Da ultima vez enviamos menos do que o esperado, verificamos se podemos sair da contingencia
                        with self.contingencia_lock:
                            if fiscalrepository.get_count_nfce_orders_to_send() == 0:
                                loggerThread.info("Saindo da contingencia")
                                self.batch_size = self.initial_batch_size
                                self.contingencia = False

                    if self.synchronous_mode and self.contingencia:
                        time.sleep(1)
                        self.process_now = True
                    else:
                        self.process_now = False
            except Exception as _:
                loggerThread.exception("Erro tratando notas em contingencia")
                self._wait(5)

    @staticmethod
    def _update_order_properties(fiscal_repository, nfce_wt_resp, order):
        xml_base64 = base64.b64encode(nfce_wt_resp)
        fiscal_repository.set_fiscal_xml_custom_property(order.pos_id, xml_base64, order.order_id)
        order_picture = OrderTaker().get_order_picture(order.pos_id, order.order_id)
        fiscal_repository.set_order_picture(order.order_id, base64.b64encode(order_picture))
        fiscal_repository.set_nfce_sent_with_xml(order.order_id, xml_base64, 1)
        return order_picture

    def _wait(self, wait_time):
        self.thread_condition.acquire()
        self.thread_condition.wait(wait_time)
        self.thread_condition.release()


class LoteException(Exception):
    def __init__(self, message, error_code):
        super(LoteException, self).__init__()
        self.message = message
        self.error_code = error_code

    def __str__(self):
        return "LoteException Error Code: {0}\\Message: {1}.".format(self.error_code, self.message)


class LoteStatus(object):
    RecebidoComSucesso = "103"
    ServicoParalisadoTemporariamente = "108"
    ServicoParalisadoSemPrevisao = "109"
    ConsumoIndevido = "656"
    ErroInterno = "999"


class LoteConsultaStatus(object):
    LoteProcessado = "104"
    LoteEmProcessamento = "105"
    ServicoParalisadoTemporariamente = "108"
    LoteComFalhaSchema = "225"
    ConsumoIndevido = "656"
    ErroInterno = "999"


class FiscalException(Exception):

    def __init__(
        self,               # type: FiscalException
        message,            # type: str
        motivo=None,        # type: Optional[str]
        response_xml=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        super(FiscalException, self).__init__()

        self.message = message
        self.motivo = motivo
        self.response_xml = response_xml

    def __str__(self):
        ret = "FiscalException: " + self.message + "."
        if self.motivo is not None:
            ret += " Motivo: " + self.motivo + "."

        return ret


class FiscalBuildException(Exception):
    def __init__(self, message, motivo=None):
        # type: (str, str) -> None
        super(FiscalBuildException, self).__init__()
        self.message = message
        self.motivo = motivo

    def __str__(self):
        ret = "FiscalBuildException: " + self.message + "."
        if self.motivo is not None:
            ret += " Motivo: " + self.motivo + "."

        return ret



# -*- coding: utf-8 -*-
"""
past.translation
==================

The ``past.translation`` package provides an import hook for Python 3 which
transparently runs ``futurize`` fixers over Python 2 code on import to convert
print statements into functions, etc.

It is intended to assist users in migrating to Python 3.x even if some
dependencies still only support Python 2.x.

Usage
-----

Once your Py2 package is installed in the usual module search path, the import
hook is invoked as follows:

    >>> from past import autotranslate
    >>> autotranslate('mypackagename')

Or:

    >>> autotranslate(['mypackage1', 'mypackage2'])

You can unregister the hook using::

    >>> from past.translation import remove_hooks
    >>> remove_hooks()

Author: Ed Schofield. 
Inspired by and based on ``uprefix`` by Vinay M. Sajip.
"""

import imp
import logging
import marshal
import os
import sys
import copy
from lib2to3.pgen2.parse import ParseError
from lib2to3.refactor import RefactoringTool

from libfuturize import fixes


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

myfixes = (list(fixes.libfuturize_fix_names_stage1) +
           list(fixes.lib2to3_fix_names_stage1) +
           list(fixes.libfuturize_fix_names_stage2) +
           list(fixes.lib2to3_fix_names_stage2))


# We detect whether the code is Py2 or Py3 by applying certain lib2to3 fixers
# to it. If the diff is empty, it's Python 3 code.

py2_detect_fixers = [
# From stage 1:
    'lib2to3.fixes.fix_apply',
    # 'lib2to3.fixes.fix_dict',        # TODO: add support for utils.viewitems() etc. and move to stage2
    'lib2to3.fixes.fix_except',
    'lib2to3.fixes.fix_execfile',
    'lib2to3.fixes.fix_exitfunc',
    'lib2to3.fixes.fix_funcattrs',
    'lib2to3.fixes.fix_filter',
    'lib2to3.fixes.fix_has_key',
    'lib2to3.fixes.fix_idioms',
    'lib2to3.fixes.fix_import',    # makes any implicit relative imports explicit. (Use with ``from __future__ import absolute_import)
    'lib2to3.fixes.fix_intern',
    'lib2to3.fixes.fix_isinstance',
    'lib2to3.fixes.fix_methodattrs',
    'lib2to3.fixes.fix_ne',
    'lib2to3.fixes.fix_numliterals',    # turns 1L into 1, 0755 into 0o755
    'lib2to3.fixes.fix_paren',
    'lib2to3.fixes.fix_print',
    'lib2to3.fixes.fix_raise',   # uses incompatible with_traceback() method on exceptions
    'lib2to3.fixes.fix_renames',
    'lib2to3.fixes.fix_reduce',
    # 'lib2to3.fixes.fix_set_literal',  # this is unnecessary and breaks Py2.6 support
    'lib2to3.fixes.fix_repr',
    'lib2to3.fixes.fix_standarderror',
    'lib2to3.fixes.fix_sys_exc',
    'lib2to3.fixes.fix_throw',
    'lib2to3.fixes.fix_tuple_params',
    'lib2to3.fixes.fix_types',
    'lib2to3.fixes.fix_ws_comma',
    'lib2to3.fixes.fix_xreadlines',

# From stage 2:
    'lib2to3.fixes.fix_basestring',
    # 'lib2to3.fixes.fix_buffer',    # perhaps not safe. Test this.
    # 'lib2to3.fixes.fix_callable',  # not needed in Py3.2+
    # 'lib2to3.fixes.fix_dict',        # TODO: add support for utils.viewitems() etc.
    'lib2to3.fixes.fix_exec',
    # 'lib2to3.fixes.fix_future',    # we don't want to remove __future__ imports
    'lib2to3.fixes.fix_getcwdu',
    # 'lib2to3.fixes.fix_imports',   # called by libfuturize.fixes.fix_future_standard_library
    # 'lib2to3.fixes.fix_imports2',  # we don't handle this yet (dbm)
    # 'lib2to3.fixes.fix_input',
    # 'lib2to3.fixes.fix_itertools',
    # 'lib2to3.fixes.fix_itertools_imports',
    'lib2to3.fixes.fix_long',
    # 'lib2to3.fixes.fix_map',
    # 'lib2to3.fixes.fix_metaclass', # causes SyntaxError in Py2! Use the one from ``six`` instead
    'lib2to3.fixes.fix_next',
    'lib2to3.fixes.fix_nonzero',     # TODO: add a decorator for mapping __bool__ to __nonzero__
    # 'lib2to3.fixes.fix_operator',    # we will need support for this by e.g. extending the Py2 operator module to provide those functions in Py3
    'lib2to3.fixes.fix_raw_input',
    # 'lib2to3.fixes.fix_unicode',   # strips off the u'' prefix, which removes a potentially helpful source of information for disambiguating unicode/byte strings
    # 'lib2to3.fixes.fix_urllib',
    'lib2to3.fixes.fix_xrange',
    # 'lib2to3.fixes.fix_zip',
]


class RTs:
    """
    A namespace for the refactoring tools. This avoids creating these at
    the module level, which slows down the module import. (See issue #117).

    There are two possible grammars: with or without the print statement.
    Hence we have two possible refactoring tool implementations.
    """
    _rt = None
    _rtp = None
    _rt_py2_detect = None
    _rtp_py2_detect = None

    @staticmethod
    def setup():
        """
        Call this before using the refactoring tools to create them on demand
        if needed.
        """
        if None in [RTs._rt, RTs._rtp]:
            RTs._rt = RefactoringTool(myfixes)
            RTs._rtp = RefactoringTool(myfixes, {'print_function': True})


    @staticmethod
    def setup_detect_python2():
        """
        Call this before using the refactoring tools to create them on demand
        if needed.
        """
        if None in [RTs._rt_py2_detect, RTs._rtp_py2_detect]:
            RTs._rt_py2_detect = RefactoringTool(py2_detect_fixers)
            RTs._rtp_py2_detect = RefactoringTool(py2_detect_fixers,
                                                  {'print_function': True})


# We need to find a prefix for the standard library, as we don't want to
# process any files there (they will already be Python 3).
#
# The following method is used by Sanjay Vinip in uprefix. This fails for
# ``conda`` environments:
#     # In a non-pythonv virtualenv, sys.real_prefix points to the installed Python.
#     # In a pythonv venv, sys.base_prefix points to the installed Python.
#     # Outside a virtual environment, sys.prefix points to the installed Python.

#     if hasattr(sys, 'real_prefix'):
#         _syslibprefix = sys.real_prefix
#     else:
#         _syslibprefix = getattr(sys, 'base_prefix', sys.prefix)

# Instead, we use the portion of the path common to both the stdlib modules
# ``math`` and ``urllib``.

def splitall(path):
    """
    Split a path into all components. From Python Cookbook.
    """
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def common_substring(s1, s2):
    """
    Returns the longest common substring to the two strings, starting from the
    left.
    """
    chunks = []
    path1 = splitall(s1)
    path2 = splitall(s2)
    for (dir1, dir2) in zip(path1, path2):
        if dir1 != dir2:
            break
        chunks.append(dir1)
    return os.path.join(*chunks)

# _stdlibprefix = common_substring(math.__file__, urllib.__file__)


def detect_python2(source, pathname):
    """
    Returns a bool indicating whether we think the code is Py2
    """
    RTs.setup_detect_python2()
    try:
        tree = RTs._rt_py2_detect.refactor_string(source, pathname)
    except ParseError as e:
        if e.msg != 'bad input' or e.value != '=':
            raise
        tree = RTs._rtp.refactor_string(source, pathname)

    if source != str(tree)[:-1]:   # remove added newline
        # The above fixers made changes, so we conclude it's Python 2 code
        logger.debug('Detected Python 2 code: {0}'.format(pathname))
        with open('/tmp/original_code.py', 'w') as f:
            f.write('### Original code (detected as py2): %s\n%s' % 
                    (pathname, source))
        with open('/tmp/py2_detection_code.py', 'w') as f:
            f.write('### Code after running py3 detection (from %s)\n%s' % 
                    (pathname, str(tree)[:-1]))
        return True
    else:
        logger.debug('Detected Python 3 code: {0}'.format(pathname))
        with open('/tmp/original_code.py', 'w') as f:
            f.write('### Original code (detected as py3): %s\n%s' % 
                    (pathname, source))
        try:
            os.remove('/tmp/futurize_code.py')
        except OSError:
            pass
        return False


class Py2Fixer(object):
    """
    An import hook class that uses lib2to3 for source-to-source translation of
    Py2 code to Py3.
    """

    # See the comments on :class:future.standard_library.RenameImport.
    # We add this attribute here so remove_hooks() and install_hooks() can
    # unambiguously detect whether the import hook is installed:
    PY2FIXER = True

    def __init__(self):
        self.found = None
        self.base_exclude_paths = ['future', 'past']
        self.exclude_paths = copy.copy(self.base_exclude_paths)
        self.include_paths = []

    def include(self, paths):
        """
        Pass in a sequence of module names such as 'plotrique.plotting' that,
        if present at the leftmost side of the full package name, would
        specify the module to be transformed from Py2 to Py3.
        """
        self.include_paths += paths

    def exclude(self, paths):
        """
        Pass in a sequence of strings such as 'mymodule' that, if
        present at the leftmost side of the full package name, would cause
        the module not to undergo any source transformation.
        """
        self.exclude_paths += paths

    def find_module(self, fullname, path=None):
        logger.debug('Running find_module: {0}...'.format(fullname))
        if '.' in fullname:
            parent, child = fullname.rsplit('.', 1)
            if path is None:
                loader = self.find_module(parent, path)
                mod = loader.load_module(parent)
                path = mod.__path__
            fullname = child

        # Perhaps we should try using the new importlib functionality in Python
        # 3.3: something like this?
        # thing = importlib.machinery.PathFinder.find_module(fullname, path)
        try:
            self.found = imp.find_module(fullname, path)
        except Exception as e:
            logger.debug('Py2Fixer could not find {0}')
            logger.debug('Exception was: {0})'.format(fullname, e))
            return None
        self.kind = self.found[-1][-1]
        if self.kind == imp.PKG_DIRECTORY:
            self.pathname = os.path.join(self.found[1], '__init__.py')
        elif self.kind == imp.PY_SOURCE:
            self.pathname = self.found[1]
        return self

    def transform(self, source):
        # This implementation uses lib2to3,
        # you can override and use something else
        # if that's better for you

        # lib2to3 likes a newline at the end
        RTs.setup()
        source += '\n'
        try:
            tree = RTs._rt.refactor_string(source, self.pathname)
        except ParseError as e:
            if e.msg != 'bad input' or e.value != '=':
                raise
            tree = RTs._rtp.refactor_string(source, self.pathname)
        # could optimise a bit for only doing str(tree) if
        # getattr(tree, 'was_changed', False) returns True
        return str(tree)[:-1] # remove added newline

    def load_module(self, fullname):
        logger.debug('Running load_module for {0}...'.format(fullname))
        if fullname in sys.modules:
            mod = sys.modules[fullname]
        else:
            if self.kind in (imp.PY_COMPILED, imp.C_EXTENSION, imp.C_BUILTIN,
                             imp.PY_FROZEN):
                convert = False
            # elif (self.pathname.startswith(_stdlibprefix)
            #       and 'site-packages' not in self.pathname):
            #     # We assume it's a stdlib package in this case. Is this too brittle?
            #     # Please file a bug report at https://github.com/PythonCharmers/python-future
            #     # if so.
            #     convert = False
            # in theory, other paths could be configured to be excluded here too
            elif any([fullname.startswith(path) for path in self.exclude_paths]):
                convert = False
            elif any([fullname.startswith(path) for path in self.include_paths]):
                convert = True
            else:
                convert = False
            if not convert:
                logger.debug('Excluded {0} from translation'.format(fullname))
                mod = imp.load_module(fullname, *self.found)
            else:
                logger.debug('Autoconverting {0} ...'.format(fullname))
                mod = imp.new_module(fullname)
                sys.modules[fullname] = mod

                # required by PEP 302
                mod.__file__ = self.pathname
                mod.__name__ = fullname
                mod.__loader__ = self

                # This:
                #     mod.__package__ = '.'.join(fullname.split('.')[:-1])
                # seems to result in "SystemError: Parent module '' not loaded,
                # cannot perform relative import" for a package's __init__.py
                # file. We use the approach below. Another option to try is the
                # minimal load_module pattern from the PEP 302 text instead.

                # Is the test in the next line more or less robust than the
                # following one? Presumably less ...
                # ispkg = self.pathname.endswith('__init__.py')
                
                if self.kind == imp.PKG_DIRECTORY:
                    mod.__path__ = [ os.path.dirname(self.pathname) ]
                    mod.__package__ = fullname
                else:
                    #else, regular module
                    mod.__path__ = []
                    mod.__package__ = fullname.rpartition('.')[0]
                    
                try:
                    cachename = imp.cache_from_source(self.pathname)
                    if not os.path.exists(cachename):
                        update_cache = True
                    else:
                        sourcetime = os.stat(self.pathname).st_mtime
                        cachetime = os.stat(cachename).st_mtime
                        update_cache = cachetime < sourcetime
                    # # Force update_cache to work around a problem with it being treated as Py3 code???
                    # update_cache = True
                    if not update_cache:
                        with open(cachename, 'rb') as f:
                            data = f.read()
                            try:
                                code = marshal.loads(data)
                            except Exception:
                                # pyc could be corrupt. Regenerate it
                                update_cache = True
                    if update_cache:
                        if self.found[0]:
                            source = self.found[0].read()
                        elif self.kind == imp.PKG_DIRECTORY:
                            with open(self.pathname) as f:
                                source = f.read()

                        if detect_python2(source, self.pathname):
                            source = self.transform(source)
                            with open('/tmp/futurized_code.py', 'w') as f:
                                f.write('### Futurized code (from %s)\n%s' % 
                                        (self.pathname, source))

                        code = compile(source, self.pathname, 'exec')

                        dirname = os.path.dirname(cachename)
                        if not os.path.exists(dirname):
                            os.makedirs(dirname)
                        try:
                            with open(cachename, 'wb') as f:
                                data = marshal.dumps(code)
                                f.write(data)
                        except Exception:   # could be write-protected
                            pass
                    exec(code, mod.__dict__)
                except Exception as e:
                    # must remove module from sys.modules
                    del sys.modules[fullname]
                    raise # keep it simple

        if self.found[0]:
            self.found[0].close()
        return mod

_hook = Py2Fixer()


def install_hooks(include_paths=(), exclude_paths=()):
    if isinstance(include_paths, str):
        include_paths = (include_paths,)
    if isinstance(exclude_paths, str):
        exclude_paths = (exclude_paths,)
    assert len(include_paths) + len(exclude_paths) > 0, 'Pass at least one argument'
    _hook.include(include_paths)
    _hook.exclude(exclude_paths)
    # _hook.debug = debug
    enable = sys.version_info[0] >= 3   # enabled for all 3.x
    if enable and _hook not in sys.meta_path:
        sys.meta_path.insert(0, _hook)  # insert at beginning. This could be made a parameter

    # We could return the hook when there are ways of configuring it
    #return _hook


def remove_hooks():
    if _hook in sys.meta_path:
        sys.meta_path.remove(_hook)


def detect_hooks():
    """
    Returns True if the import hooks are installed, False if not.
    """
    return _hook in sys.meta_path
    # present = any([hasattr(hook, 'PY2FIXER') for hook in sys.meta_path])
    # return present


class hooks(object):
    """
    Acts as a context manager. Use like this:
    
    >>> from past import translation
    >>> with translation.hooks():
    ...     import mypy2module
    >>> import requests        # py2/3 compatible anyway
    >>> # etc.
    """
    def __enter__(self):
        self.hooks_were_installed = detect_hooks()
        install_hooks()
        return self

    def __exit__(self, *args):
        if not self.hooks_were_installed:
            remove_hooks()


class suspend_hooks(object):
    """
    Acts as a context manager. Use like this:
    
    >>> from past import translation
    >>> translation.install_hooks()
    >>> import http.client
    >>> # ...
    >>> with translation.suspend_hooks():
    >>>     import requests     # or others that support Py2/3

    If the hooks were disabled before the context, they are not installed when
    the context is left.
    """
    def __enter__(self):
        self.hooks_were_installed = detect_hooks()
        remove_hooks()
        return self
    def __exit__(self, *args):
        if self.hooks_were_installed:
            install_hooks()




from typing import Optional

MENU_MANAGER_AUTHENTICATED_USER = {}


def get_menu_manager_authenticated_user(
    pos_id,     # type: int
):
    # type: (...) -> Optional[str]

    if pos_id not in MENU_MANAGER_AUTHENTICATED_USER:
        return None

    return MENU_MANAGER_AUTHENTICATED_USER[pos_id]


def set_menu_manager_authenticated_user(
    pos_id,     # type: int
    user_id     # type: str
):
    # type: (...) -> None

    global MENU_MANAGER_AUTHENTICATED_USER
    MENU_MANAGER_AUTHENTICATED_USER[pos_id] = user_id


def clear_menu_manager_authenticated_user(
    pos_id,     # type: int
):
    # type: (...) -> None

    global MENU_MANAGER_AUTHENTICATED_USER

    if pos_id not in MENU_MANAGER_AUTHENTICATED_USER:
        return

    del MENU_MANAGER_AUTHENTICATED_USER[pos_id]



""" Image format handling class

This module contains the image format handler :py:class:`EscposImage`.

:author: `Michael Billington <michael.billington@gmail.com>`_
:organization: `python-escpos <https://github.com/python-escpos>`_
:copyright: Copyright (c) 2016 Michael Billington <michael.billington@gmail.com>
:license: GNU GPL v3
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math
from PIL import Image, ImageOps


class EscposImage(object):
    """
    Load images in, and output ESC/POS formats.

    The class is designed to efficiently delegate image processing to
    PIL, rather than spend CPU cycles looping over pixels.
    """

    def __init__(self, img_source):
        """
        Load in an image
        
        :param img_source: PIL.Image, or filename to load one from.
        """
        if isinstance(img_source, Image.Image):
            img_original = img_source
        else:
            img_original = Image.open(img_source)

        # store image for eventual further processing (splitting)
        self.img_original = img_original

        # Convert to white RGB background, paste over white background
        # to strip alpha.
        img_original = img_original.convert('RGBA')
        im = Image.new("RGB", img_original.size, (255, 255, 255))
        im.paste(img_original, mask=img_original.split()[3])
        # Convert down to greyscale
        im = im.convert("L") 
        # Invert: Only works on 'L' images
        im = ImageOps.invert(im)
        # Pure black and white
        self._im = im.convert("1")
    
    @property
    def width(self):
        """
        Width of image in pixels
        """
        width_pixels, _ = self._im.size
        return width_pixels

    @property
    def width_bytes(self):
        """
        Width of image if you use 8 pixels per byte and 0-pad at the end.
        """
        return (self.width + 7) >> 3

    @property
    def height(self):
        """
        Height of image in pixels
        """
        _, height_pixels = self._im.size
        return height_pixels

    def to_column_format(self, high_density_vertical=True):
        """
        Extract slices of an image as equal-sized blobs of column-format data.

        :param high_density_vertical: Printed line height in dots
        """
        im = self._im.transpose(Image.ROTATE_270).transpose(Image.FLIP_LEFT_RIGHT)
        line_height = 24 if high_density_vertical else 8
        width_pixels, height_pixels = im.size
        top = 0
        left = 0
        while left < width_pixels:
            box = (left, top, left + line_height, top + height_pixels)
            im_slice = im.transform((line_height, height_pixels), Image.EXTENT, box)
            im_bytes = im_slice.tobytes()
            yield(im_bytes)
            left += line_height

    def to_raster_format(self):
        """
        Convert image to raster-format binary
        """
        return self._im.tobytes()

    def split(self, fragment_height):
        """
        Split an image into multiple fragments after fragment_height pixels

        :param fragment_height: height of fragment
        :return: list of PIL objects
        """
        passes = int(math.ceil(self.height/fragment_height))
        fragments = []
        for n in range(0, passes):
            left = 0
            right = self.width
            upper = n * fragment_height
            lower = min((n + 1) * fragment_height, self.height)
            box = (left, upper, right, lower)
            fragments.append(self.img_original.crop(box))
        return fragments



import os

import cfgtools

delivery_store_pickup_config = None


def should_ask_delivery_type():
    # type: (...) -> bool

    global delivery_store_pickup_config

    if delivery_store_pickup_config is None:
        config = cfgtools.read(os.environ["LOADERCFG"])
        delivery_store_pickup_config = config.find_value("Customizations.ManualDeliveryStorePickup", "false")

    return delivery_store_pickup_config.lower() == "true"



# tin.py - functions for handling South Africa Tax Reference Number
# coding: utf-8
#
# Copyright (C) 2019 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""TIN (South African Tax Identification Number).

The South African Tax Identification Number (TIN or Tax Reference Number) is
issued to individuals and legal entities for tax purposes. The number
consists of 10 digits.

More information:

* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/South-Africa-TIN.pdf
* https://www.sars.gov.za/

>>> validate('0001339050')
'0001339050'
>>> validate('2449/494/16/0')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('9125568')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('084308984-8')
'0843089848'
"""  # noqa: E501

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    return clean(number, ' -/').upper().strip()


def validate(number):
    """Check if the number is a valid South Africa Tax Reference Number.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if len(number) != 10:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[0] not in '01239':
        raise InvalidComponent()
    return luhn.validate(number)


def is_valid(number):
    """Check if the number is a valid South Africa Tax Reference Number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



from threading import local

_local = local()


def get_current_context(silent=False):
    """Returns the current click context.  This can be used as a way to
    access the current context object from anywhere.  This is a more implicit
    alternative to the :func:`pass_context` decorator.  This function is
    primarily useful for helpers such as :func:`echo` which might be
    interested in changing its behavior based on the current context.

    To push the current context, :meth:`Context.scope` can be used.

    .. versionadded:: 5.0

    :param silent: if set to `True` the return value is `None` if no context
                   is available.  The default behavior is to raise a
                   :exc:`RuntimeError`.
    """
    try:
        return _local.stack[-1]
    except (AttributeError, IndexError):
        if not silent:
            raise RuntimeError("There is no active click context.")


def push_context(ctx):
    """Pushes a new context to the current stack."""
    _local.__dict__.setdefault("stack", []).append(ctx)


def pop_context():
    """Removes the top level from the stack."""
    _local.stack.pop()


def resolve_color_default(color=None):
    """"Internal helper to get the default value of the color flag.  If a
    value is passed it's returned unchanged, otherwise it's looked up from
    the current context.
    """
    if color is not None:
        return color
    ctx = get_current_context(silent=True)
    if ctx is not None:
        return ctx.color



"""Fix UserDict.

Incomplete!

TODO: base this on fix_urllib perhaps?
"""


# Local imports
from lib2to3 import fixer_base
from lib2to3.fixer_util import Name, attr_chain
from lib2to3.fixes.fix_imports import alternates, build_pattern, FixImports

MAPPING = {'UserDict':  'collections',
}

# def alternates(members):
#     return "(" + "|".join(map(repr, members)) + ")"
# 
# 
# def build_pattern(mapping=MAPPING):
#     mod_list = ' | '.join(["module_name='%s'" % key for key in mapping])
#     bare_names = alternates(mapping.keys())
# 
#     yield """name_import=import_name< 'import' ((%s) |
#                multiple_imports=dotted_as_names< any* (%s) any* >) >
#           """ % (mod_list, mod_list)
#     yield """import_from< 'from' (%s) 'import' ['(']
#               ( any | import_as_name< any 'as' any > |
#                 import_as_names< any* >)  [')'] >
#           """ % mod_list
#     yield """import_name< 'import' (dotted_as_name< (%s) 'as' any > |
#                multiple_imports=dotted_as_names<
#                  any* dotted_as_name< (%s) 'as' any > any* >) >
#           """ % (mod_list, mod_list)
# 
#     # Find usages of module members in code e.g. thread.foo(bar)
#     yield "power< bare_with_attr=(%s) trailer<'.' any > any* >" % bare_names


# class FixUserDict(fixer_base.BaseFix):
class FixUserdict(FixImports):

    BM_compatible = True
    keep_line_order = True
    # This is overridden in fix_imports2.
    mapping = MAPPING

    # We want to run this fixer late, so fix_import doesn't try to make stdlib
    # renames into relative imports.
    run_order = 6

    def build_pattern(self):
        return "|".join(build_pattern(self.mapping))

    def compile_pattern(self):
        # We override this, so MAPPING can be pragmatically altered and the
        # changes will be reflected in PATTERN.
        self.PATTERN = self.build_pattern()
        super(FixImports, self).compile_pattern()

    # Don't match the node if it's within another match.
    def match(self, node):
        match = super(FixImports, self).match
        results = match(node)
        if results:
            # Module usage could be in the trailer of an attribute lookup, so we
            # might have nested matches when "bare_with_attr" is present.
            if "bare_with_attr" not in results and \
                    any(match(obj) for obj in attr_chain(node, "parent")):
                return False
            return results
        return False

    def start_tree(self, tree, filename):
        super(FixImports, self).start_tree(tree, filename)
        self.replace = {}

    def transform(self, node, results):
        import_mod = results.get("module_name")
        if import_mod:
            mod_name = import_mod.value
            new_name = unicode(self.mapping[mod_name])
            import_mod.replace(Name(new_name, prefix=import_mod.prefix))
            if "name_import" in results:
                # If it's not a "from x import x, y" or "import x as y" import,
                # marked its usage to be replaced.
                self.replace[mod_name] = new_name
            if "multiple_imports" in results:
                # This is a nasty hack to fix multiple imports on a line (e.g.,
                # "import StringIO, urlparse"). The problem is that I can't
                # figure out an easy way to make a pattern recognize the keys of
                # MAPPING randomly sprinkled in an import statement.
                results = self.match(node)
                if results:
                    self.transform(node, results)
        else:
            # Replace usage of the module.
            bare_name = results["bare_with_attr"][0]
            new_name = self.replace.get(bare_name.value)
            if new_name:
                bare_name.replace(Name(new_name, prefix=bare_name.prefix))




# nit.py - functions for handling Guatemala NIT numbers
# coding: utf-8
#
# Copyright (C) 2019 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIT (Número de Identificación Tributaria, Guatemala tax number).

The Número de Identificación Tributaria (NIT) is an identifier of legal
entities for tax purposes in Guatemala.

The number consists of 2 to 12 characters, where the last one is the check
digit (a digit or the letter K) and the rest are digits. Leading zeroes are
usually omitted. Digits and check digit are usually separated with a hyphen.

More information:

* https://portal.sat.gob.gt/portal/descarga/6524/factura-electronica-fel/25542/fel-reglas-y-validaciones.pdf (page 58)
* https://portal.sat.gob.gt/portal/consulta-cui-nit/

>>> validate('576937-K')
'576937K'
>>> validate('7108-0')
'71080'
>>> validate('8977112-0')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1234567890123')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('39525503')
'3952550-3'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').upper().strip().lstrip('0')


def calc_check_digit(number):
    """Calculate the check digit. The number passed should not have the
    check digit included."""
    c = -sum(i * int(n) for i, n in enumerate(reversed(number), 2)) % 11
    return 'K' if c == 10 else str(c)


def validate(number):
    """Check if the number is a valid Guatemala NIT number.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if len(number) < 2 or len(number) > 12:
        raise InvalidLength()
    if not isdigits(number[:-1]):
        raise InvalidFormat()
    if number[-1] != 'K' and not isdigits(number[-1]):
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number[:-1]):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid Guatemala NIT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join([number[:-1], number[-1]])



from utils import update_tenders
from create_order import create_order
from must_sync_ui import (
    must_sync_ui,
    sync_ui_and_model,
)
from validate_status_order import order_state_is_valid
from order_was_already_saved import order_was_already_saved
from close_active_order_and_pause_user import close_active_order_and_pause_user



from typing import Optional, Dict


class Event(object):
    def __init__(self, subject, event_type=None, data=None, source=0, synchronous=False, timeout=-1, headers=None):
        # type: (str, Optional[str], Optional[bytes], int, bool, int, Optional[Dict[str, str]]) -> None
        self.subject = subject
        self.event_type = event_type
        self.source = source
        self.data = data
        self.synchronous = synchronous
        self.timeout = timeout
        self.headers = headers if headers is not None else {}
        self.imp_message = None

    def __str__(self):
        # type: () -> str
        return str({
            "subject": self.subject,
            "event_type": self.event_type,
            "synchronous": self.synchronous,
            "timeout": self.timeout,
            "headers": self.headers
        })



# -*- coding: utf-8 -*-
from typing import Dict, Any
from stdnum import ean
from xml.sax.saxutils import escape

from helper import unicode_2_ascii
from old_helper import round_half_away_from_zero

from pos_model import Order, TaxItem, SaleItem
from common import FiscalParameterController
from pos_util import SaleLineUtil

from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _ContextKeys import ContextKeys
from _TaxBuilder import TaxBuilderLucroReal, TaxBuilderSimplesNacional, TaxBuilderSimplesNacionalSublimiteReceita


class DetBuilder(NfceXmlPartBuilder):
    def __init__(self, ambiente, crt, fiscal_parameter_controller, sale_line_util, versao_ws):
        # type: (unicode, int, FiscalParameterController, SaleLineUtil, int) -> None

        self.versao_ws = versao_ws
        self.ambiente = ambiente
        self.crt = crt
        self.fiscal_parameter_controller = fiscal_parameter_controller
        self.sale_line_util = sale_line_util

        if self.crt == 1:
            self.tax_builder = TaxBuilderSimplesNacional()
        elif self.crt == 2:
            self.tax_builder = TaxBuilderSimplesNacionalSublimiteReceita()
        elif self.crt == 3:
            self.tax_builder = TaxBuilderLucroReal()
        else:
            raise Exception("CRT Não informado")

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        item_no = context[ContextKeys.item_no]  # type: int
        sale_item = context[ContextKeys.current_sale_item]  # type: SaleItem

        product_code = sale_item.part_code
        cest = self.fiscal_parameter_controller.get_optional_parameter(product_code, "CEST")
        cBenef = self.fiscal_parameter_controller.get_optional_parameter(product_code, "cBenef")

        ncm = self.fiscal_parameter_controller.get_parameter(product_code, "NCM")
        if ncm is None:
            raise Exception("NCM Não encontrado\\%s-%s" % (product_code, sale_item.product_name))

        cfop = self.fiscal_parameter_controller.get_parameter(product_code, "CFOP")
        if cfop is None:
            raise Exception("CFOP Não encontrado\\%s-%s" % (product_code, sale_item.product_name))

        csosn = None
        if self.crt == 1:
            csosn = self.fiscal_parameter_controller.get_parameter(product_code, "CSOSN")
            if csosn is None:
                raise Exception("CSOSN Não encontrado\\%s-%s" % (product_code, sale_item.product_name))
            csosn = int(csosn)

        cst_icms = self.fiscal_parameter_controller.get_parameter(product_code, "CST_ICMS")
        if cst_icms is None:
            raise Exception("CST ICMS Não encontrado\\%s-%s" % (product_code, sale_item.product_name))
        cst_icms = int(cst_icms)

        cst_pis = self.fiscal_parameter_controller.get_parameter(product_code, "CST_PIS")
        if cst_pis is None:
            raise Exception("CST PIS Não encontrado\\%s-%s" % (product_code, sale_item.product_name))
        cst_pis = int(cst_pis)

        cst_cofins = self.fiscal_parameter_controller.get_parameter(product_code, "CST_COFINS")
        if cst_cofins is None:
            raise Exception("CST COFINS Não encontrado\\%s-%s" % (product_code, sale_item.product_name))
        cst_cofins = int(cst_cofins)

        base_reduction = float(self.fiscal_parameter_controller.get_optional_parameter(product_code, "BASE_REDUCTION") or 0)
        cst_reason = self.fiscal_parameter_controller.get_optional_parameter(product_code, "CST_REASON")

        if self.ambiente == '1':
            product_name = unicode_2_ascii(escape(sale_item.product_name).strip())
        else:
            product_name = "NOTA FISCAL EMITIDA EM AMBIENTE DE HOMOLOGACAO - SEM VALOR FISCAL"

        cest_xml = "" if cest is None else "<CEST>{cest:>07}</CEST>".format(cest=cest)
        c_benef = "" if cBenef is None or cst_icms == 0 else "<cBenef>{cBenef}</cBenef>".format(cBenef=cBenef)

        prod_xml_template = "<prod>" \
                            "<cProd>{c_prod}</cProd>" \
                            "{c_ean}" \
                            "<xProd>{x_prod}</xProd>" \
                            "<NCM>{ncm:0>8}</NCM>" \
                            "{cest_xml}" \
                            "{c_benef}" \
                            "<CFOP>{cfop}</CFOP>" \
                            "<uCom>{measure_unit}</uCom>" \
                            "<qCom>{q_com:0.4f}</qCom>" \
                            "<vUnCom>{v_un_com:0.3f}</vUnCom>" \
                            "<vProd>{v_prod:.02f}</vProd>" \
                            "{c_ean_trib}" \
                            "<uTrib>{u_trib}</uTrib>" \
                            "<qTrib>{q_trib}</qTrib>" \
                            "<vUnTrib>{v_un_trib}</vUnTrib>" \
                            "{discount_xml}" \
                            "<indTot>{ind_tot}</indTot>" \
                            "</prod>"

        item_discount = sale_item.item_discount or 0
        discount_xml = ""
        discount_value = 0 if item_discount < 0 else item_discount
        increase_value = 0 if item_discount > 0 else abs(item_discount)

        if discount_value > 0:
            discount_xml += "<{0}>{1:.02f}</{0}>".format("vDesc",  abs(discount_value))
        if increase_value > 0:
            discount_xml += "<{0}>{1:.02f}</{0}>".format("vOutro", abs(increase_value))

        item_value = round_half_away_from_zero(sale_item.item_price / sale_item.quantity, 2)

        c_ean = "SEM GTIN"
        if sale_item.ean and ean.is_valid(number=sale_item.ean):
            c_ean = sale_item.ean

        prod_xml = prod_xml_template.format(
            c_prod=product_code,
            c_ean="<cEAN>{}</cEAN>".format(c_ean),
            x_prod=product_name,
            ncm=int(ncm),
            cest_xml=cest_xml,
            c_benef=c_benef,
            cfop=cfop,
            q_com=sale_item.quantity,
            v_un_com=item_value,
            v_prod=sale_item.item_price,
            c_ean_trib="<cEANTrib>{}</cEANTrib>".format(c_ean),
            u_trib="UN",
            q_trib=sale_item.quantity,
            v_un_trib=item_value,
            discount_xml=discount_xml,
            ind_tot=1,
            measure_unit=sale_item.measure_unit.upper()
        )

        parcial_imposto_atual = 0
        impost_xml_template = "<imposto><vTotTrib>{v_tot_trib:0.2f}</vTotTrib>{tax_xml}</imposto>"

        fcp_tax_item = sale_item.get_tax_item("FCP")  # type: TaxItem
        icms_tax_item = sale_item.get_tax_item("ICMS")  # type: TaxItem
        xml, icms_vbc, icms_tax, deson, fcp_tax, fcp_st, fcp_ant = self.tax_builder.calculate_icms(cst_icms, csosn, icms_tax_item, fcp_tax_item, base_reduction, cst_reason, product_code)

        tax_xml = xml
        parcial_icms_vbc = icms_vbc
        parcial_imposto_atual += icms_tax

        pis_tax_item = sale_item.get_tax_item("PIS")  # type: TaxItem
        xml, vbc, pis_tax = self.tax_builder.calculate_pis(cst_pis, pis_tax_item)
        tax_xml += xml
        parcial_imposto_atual += pis_tax

        cofins_tax_item = sale_item.get_tax_item("COFINS")  # type: TaxItem
        xml, vbc, cofins_tax = self.tax_builder.calculate_cofins(cst_cofins, cofins_tax_item)
        tax_xml += xml
        parcial_imposto_atual += cofins_tax

        imposto_xml = impost_xml_template.format(v_tot_trib=parcial_imposto_atual, tax_xml=tax_xml)

        xml = "<det nItem=\"{0}\">{prod_xml}{imposto_xml}</det>".format(
            item_no,
            prod_xml=prod_xml,
            imposto_xml=imposto_xml
        )

        context[ContextKeys.total_cofins] += cofins_tax
        context[ContextKeys.total_icms] += icms_tax
        context[ContextKeys.total_pis] += pis_tax
        context[ContextKeys.total_prod] += sale_item.item_price
        context[ContextKeys.total_icms_vbc] += parcial_icms_vbc

        if ContextKeys.deson not in context:
            context[ContextKeys.deson] = 0
        context[ContextKeys.deson] += deson

        if ContextKeys.total_fcp not in context:
            context[ContextKeys.total_fcp] = 0
        context[ContextKeys.total_fcp] += fcp_tax

        if ContextKeys.total_fcp_st not in context:
            context[ContextKeys.total_fcp_st] = 0
        context[ContextKeys.total_fcp_st] += fcp_st

        if ContextKeys.total_fcp_ant not in context:
            context[ContextKeys.total_fcp_ant] = 0
        context[ContextKeys.total_fcp_ant] += fcp_ant

        return xml



from logging import Logger
from werkzeug.serving import make_server
from threading import Thread

from flask import Flask, Request, Response, request
from simplehttprouter import SimpleHttpRouter, Route
from typing import Dict, Callable, Optional, Any


class FlaskAdapter(object):
    def __init__(self, routes, logger=None):
        # type: (Dict[Route, Callable[[Request], Response]], Optional[Logger]) -> None
        self.routes = routes
        self.logger = logger

        self.app = None
        self.router = SimpleHttpRouter(list(routes.keys()))
        self.server_thread = None
        self.srv = None
        self.ctx = None

    def start(self, name, host, port):
        # type: (str, str, int) -> Any
        self.app = Flask(name)
        self.app.add_url_rule(
            "/", "catch_all",
            self.handle_messages,
            defaults={"path": "/"},
            methods=["GET", "POST", "PUT", "OPTIONS", "HEAD", "DELETE"]
        )
        self.app.add_url_rule(
            "/<path:path>",
            "catch_all",
            self.handle_messages,
            methods=["GET", "POST", "PUT", "OPTIONS", "HEAD", "DELETE"]
        )
        if host is not None:
            self.server_thread = Thread(target=self.run_server, args=[host, port])
            self.server_thread.daemon = True
            self.server_thread.start()

        return self.app

    def stop(self):
        self.srv.shutdown()

    def run_server(self, host, port):
        self.srv = make_server(host, port, self.app, True)
        self.ctx = self.app.app_context()
        self.ctx.push()
        self.srv.serve_forever()

    def handle_messages(self, path):
        # type: (str) -> Response
        try:
            path = "/" + path if path != "/" else path
            method = request.method

            route = Route(method, path)
            found_route = self.router.get_route(route)
            path_parameters = self.router.get_path_parameters(route)
            if found_route is None:
                if self.logger:
                    self.logger.exception("No Route Found - Returning 404")
                return Response(status=404)

            func = self.routes[found_route]
            return func(path_parameters, request)

        except Exception as _:
            if self.logger:
                self.logger.exception("Error Handling Request - Returning 500")
            return Response(status=500)



import json

from messagebus import (
    AckMessage,
    Event,
    Message,
    MessageBus,
    NakMessage,
)
from paymentrouter import PaymentRouter
from typing import Any, Optional
from util import Cache

from ._CommonProcessor import _CommonProcessor


class CacheProcessor(_CommonProcessor):
    def __init__(
        self,
        payment_router,     # type: PaymentRouter
        cache,              # type: Cache
    ):
        # type: (...) -> None

        super(CacheProcessor, self).__init__(payment_router)

        self.cache = cache

    def parse_data(
        self,
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> str

        return data

    def call_business(
        self,
        data,   # type: str
    ):
        # type: (...) -> Optional[str]

        data = json.loads(data)

        eft_processor = data.get("eftProcessor")
        receipt_data = data.get("receiptData", None)

        if receipt_data is not None:
            self.cache.set_last_receipt(
                receipt_data=receipt_data,
                eft_processor=eft_processor,
            )
            return None

        return self.payment_router.get_last_receipt(eft_processor=eft_processor)

    def format_response(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        input_obj,      # type: Any
        result,         # type: Any
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=AckMessage(data=result),
        )

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        input_obj,      # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        self.payment_router._logger.exception("exception: {}".format(exception))
        message_bus.reply_message(
            message=message,
            reply_message=NakMessage(data=exception.message),
        )



from ._PosUser import PosUser
from ._UpdateType import UpdateType
from ._UpdateControl import UpdateControl
from ._PosVersions import PosVersions
from ._BaseThread import BaseThread
from ._ListenedTokens import ListenedTokens
from ._ListenedEvents import ListenedEvents
from ._UpdateData import UpdateData
from ._MediaUpdateData import MediaUpdateData
from ._UserUpdateData import UserUpdateData
from ._UpdateStatus import UpdateStatus
from ._Payment import Payment
from ._PosSaleTypeNavigation import PosSaleTypeNavigation
from ._OperatorStatus import OperatorStatus
from ._UpdateStep import UpdateStep
from ._AsyncUuidMessageProcessor import AsyncUuidMessageProcessor
from ._UpdateMediaJson import UpdateMediaJson
from ._Updater import Updater
from ._ModelJsonEncoder import ModelJsonEncoder



# tn.py - functions for handling Egypt Tax Number numbers
# coding: utf-8
#
# Copyright (C) 2022 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

u"""Tax Registration Number (الرقم الضريبي, Egypt tax number).

This number consists of 9 digits, usually separated into three groups
using hyphens to make it easier to read, like XXX-XXX-XXX.

More information:

* https://emsp.mts.gov.eg:8181/EMDB-web/faces/authoritiesandcompanies/authority/website/SearchAuthority.xhtml?lang=en

>>> validate('100-531-385')
'100531385'
>>> validate(u'٣٣١-١٠٥-٢٦٨')
'331105268'
>>> validate('12345')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('VV3456789')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> format('100531385')
'100-531-385'
"""  # noqa: E501

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


_ARABIC_NUMBERS_MAP = {
    # Arabic-indic digits.
    u'٠': '0',
    u'١': '1',
    u'٢': '2',
    u'٣': '3',
    u'٤': '4',
    u'٥': '5',
    u'٦': '6',
    u'٧': '7',
    u'٨': '8',
    u'٩': '9',
    # Extended arabic-indic digits.
    u'۰': '0',
    u'۱': '1',
    u'۲': '2',
    u'۳': '3',
    u'۴': '4',
    u'۵': '5',
    u'۶': '6',
    u'۷': '7',
    u'۸': '8',
    u'۹': '9',
}


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace. It also converts arabic numbers.
    """
    try:
        return str(''.join((_ARABIC_NUMBERS_MAP.get(c, c) for c in clean(number, ' -/').strip())))
    except UnicodeError:  # pragma: no cover (Python 2 specific)
        raise InvalidFormat()


def validate(number):
    """Check if the number is a valid Egypt Tax Number number.

    This checks the length and formatting.
    """
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    return number


def is_valid(number):
    """Check if the number is a valid Egypt Tax Number number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join([number[:3], number[3:-3], number[-3:]])



# -*- coding: utf-8 -*-

import json

import sysactions
from actions import mb_context
from persistence import Driver


@sysactions.action
def get_kiosk_prices(pos_id):
    conn = None
    params = {
        'pricelist': 'EI'
    }
    try:
        conn = Driver().open(mb_context, pos_id, service_name="Persistence")
        cursor = conn.pselect("getProductPrices", **params)
        products = {row.get_entry(0): {
            'default': float(row.get_entry(1) or 0),
            'added': float(row.get_entry(2) or 0)
        } for row in cursor}

        combos = {}
        cursor = conn.pselect("getCompositions")
        for row in cursor:
            combo_code = row.get_entry(0)
            part_code = row.get_entry(1)
            default_qty = int(row.get_entry(2))
            min_qty = int(row.get_entry(3) or 0)
            part_is_option = int(row.get_entry(4)) == 1
            item_code = row.get_entry(5)
            if part_is_option:
                current_value = _get_best_price(products, "1", combo_code, part_code, item_code.split(','))
            else:
                current_value = _get_best_price(products, "1", combo_code, part_code, [part_code])

            value = current_value * (default_qty or min_qty or 1)

            if not combos.get(combo_code):
                combos[combo_code] = {'default': 0}
            combos[combo_code]['default'] += value

        products.update(combos)
        return json.dumps(products)
    finally:
        if conn:
            conn.close()


def _get_best_price(products, cont, combo_code, part_code, items_list):
    best_price = 99999
    for item_to_find in items_list:
        current_item_price = (
                    products.get('{}.{}.{}.{}'.format(cont, combo_code, part_code, item_to_find)) or
                    products.get('{}.{}.{}'.format(combo_code, part_code, item_to_find)) or
                    products.get('{}.{}'.format(combo_code, item_to_find)) or
                    products.get('{}.{}'.format(part_code, item_to_find)) or
                    products.get('{}.{}'.format(cont, item_to_find)) or
                    products.get(item_to_find) or {}).get('default', 0)
        if 0 < current_item_price < best_price:
            best_price = current_item_price
    if best_price != 99999:
        return best_price
    return 0



import logging

from datetime import datetime
from logging import Logger
from typing import Optional

from cfgtools import read
from helper import config_logger

from services.model import (
    Configs,
    AlertsConfigurations,
)


class Configurator(object):

    def __init__(
        self,
        loader_path,            # type: str
        databases_path,         # type: str
        backup_databases_path,  # type: str
    ):
        # type: (...) -> None

        self.config = Configs()
        self.config.loader_path = loader_path
        self.config.databases_path = databases_path
        self.config.backup_databases_path = backup_databases_path
        self.reader = read(file_or_path=loader_path)
        self.group = "Alerts"
        self.logger = None

    def read_loader_configs(
        self,
    ):
        # type: (...) -> Configs

        self.configure_loggers()
        self.logger.info("Start read_loader_configs")
        self.config.component_name = self.reader.find_value("Identification.ComponentName")

        maintenance_window_start = self._get_config(
            path="Alerts.Maintenance.Window.StartTime",
        )
        maintenance_window_end = self._get_config(
            path="Alerts.Maintenance.Window.EndTime",
        )

        if maintenance_window_start and maintenance_window_end:
            maintenance_window_start = datetime.strptime(maintenance_window_start, "%H:%M:%S").time()
            maintenance_window_end = datetime.strptime(maintenance_window_end, "%H:%M:%S").time()

        time_calling = self._get_config(
            path="Alerts.TimeCalling",
            default_value=1
        )
        configured_alerts_endpoint = self._get_config(
            path="Alerts.EndPoints.ConfiguredAlerts",
            default_value=""
        )
        days_to_keep_data = self._get_config(
            path="Alerts.Maintenance.DaysToKeepData",
            default_value=45
        )
        self.config.alerts = AlertsConfigurations(
            time_calling=int(time_calling),
            configured_alerts_endpoint=str(configured_alerts_endpoint),
            maintenance_window_start=maintenance_window_start,
            maintenance_window_end=maintenance_window_end,
            days_to_keep_data=int(days_to_keep_data),
        )

        self.logger.info("End read_loader_configs")

        return self.config

    def configure_loggers(
        self,
    ):
        # type: (...) -> None

        config_logger(
            loader_path=self.config.loader_path,
            log_name="Alerts",
            max_size=10485760,
            max_files=15,
        )
        self.logger = logging.getLogger(name="Alerts")

    def get_loggers(
        self,
    ):
        # type: (...) -> Logger

        return self.logger

    def _get_config(
        self,
        path,                       # type: str
        default_value=None,         # type: Optional[bool]
        multiple_values=False,      # type: Optional[bool]
    ):
        # type: (...) -> str

        if multiple_values:
            values = self.reader.find_values(
                path=path,
            )
            return values or default_value

        return self.reader.find_value(
            path=path,
            default=default_value,
        )



import time
import json
from datetime import datetime
from logging import Logger
from threading import Thread

from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Event

from externals.repository import AlertsRepository


class RetrySentEventService(Thread):
    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        alerts_repository,  # type: AlertsRepository
        time_to_call,       # type: int
        logger,             # type: Logger
    ):
        # type: (...) -> None

        super(RetrySentEventService, self).__init__()
        self.running = True
        self.daemon = True
        self.message_bus = message_bus
        self.time_to_call = time_to_call
        self.alerts_repository = alerts_repository
        self.logger = logger
        self.event_action = "PosAlertNotification"

    def _sleep(
        self,
    ):
        # type: (...) -> None

        time.sleep(self.time_to_call * 60)

    def send_event(
        self,
        event_data    # type: str
    ):
        # type: (...) -> None

        event_to_send = Event(
            subject=self.event_action,
            event_type="",
            data=event_data,
        )
        self.message_bus.publish_event(event=event_to_send)

    def handle_events_repository(
        self,
    ):
        # type: (...) -> None

        events = self.alerts_repository.get_events_with_sent_status(sent_status=0)
        for event in events:
            event['notification_json'] = json.loads(event.get('notification_json', {}))

            try:
                event_time = datetime.strptime(event.get('time_stamp'), '%Y-%m-%d %H:%M:%S.%f')
            except ValueError:
                event_time = datetime.strptime(event.get('time_stamp'), '%Y-%m-%d %H:%M:%S')

            fade_time = int(event.get('notification_json').get('alertJSON', {}).get('fadeTime', 0))
            if (datetime.utcnow() - event_time).total_seconds() > fade_time:
                self.alerts_repository.set_sent_status(
                    event_id=event.get('id'),
                    sent_status=-1,
                )
            else:
                event_data = json.dumps(obj=event.get('notification_json'))
                self.logger.info("Sending event with event_id: {} and data: {}".format(event.get('id'), event_data))
                self.send_event(event_data=event_data)

    def run(
        self,
    ):
        while self.running:
            try:
                self.handle_events_repository()
            except Exception as ex:
                self.logger.warning("Error in RetrySentEventService", exc_info=ex)
            finally:
                self._sleep()



from application.domain import PrintReport
from sysactions import get_model, print_report


class MsbBusPrintReport(PrintReport):
    def send_print(self, pos_id, preview, report_name, *report_params):
        model = get_model(pos_id)
        return print_report(pos_id, model, preview, report_name, *report_params)



from sys import version_info

if version_info[0] <= 2:
    int2oct = chr
    ints2octs = lambda s: ''.join([ int2oct(x) for x in s ])
    null = ''
    oct2int = ord
    octs2ints = lambda s: [ oct2int(x) for x in s ]
    str2octs = lambda x: x
    octs2str = lambda x: x
    isOctetsType = lambda s: isinstance(s, str)
    isStringType = lambda s: isinstance(s, (str, unicode))
else:
    ints2octs = bytes
    int2oct = lambda x: ints2octs((x,))
    null = ints2octs()
    oct2int = lambda x: x
    octs2ints = lambda s: [ x for x in s ]
    str2octs = lambda x: x.encode()
    octs2str = lambda x: x.decode()
    isOctetsType = lambda s: isinstance(s, bytes)
    isStringType = lambda s: isinstance(s, str)



from datetime import datetime


class DeviceStatusHistory(object):
    def __init__(
        self,
        enabled,    # type: bool
        operator,   # type: str
        timestamp,  # type: datetime
    ):
        # type: (...) -> None

        self.timestamp = timestamp
        self.operator = operator
        self.enabled = enabled
